# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dowhy',
 'dowhy.api',
 'dowhy.causal_estimators',
 'dowhy.causal_identifier',
 'dowhy.causal_refuters',
 'dowhy.data_transformers',
 'dowhy.do_samplers',
 'dowhy.gcm',
 'dowhy.gcm.independence_test',
 'dowhy.gcm.ml',
 'dowhy.gcm.util',
 'dowhy.graph_learners',
 'dowhy.interpreters',
 'dowhy.utils',
 'dowhy.utils.dgps']

package_data = \
{'': ['*']}

install_requires = \
['causal-learn>=0.1.3.0,<0.2.0.0',
 'joblib>=1.1.0,<2.0.0',
 'networkx>=2.8.5,<3.0.0',
 'numpy>=1.23.1,<2.0.0',
 'pandas>=1.4.3,<2.0.0',
 'scikit-learn==1.0.2',
 'scipy>=1.8.1,<2.0.0',
 'sphinx_design>=0.3.0,<0.4.0',
 'statsmodels>=0.13.2,<0.14.0',
 'sympy>=1.10.1,<2.0.0',
 'tqdm>=4.64.0,<5.0.0']

extras_require = \
{':extra == "causalml"': ['cython>=0.29.32,<0.30.0'],
 'causalml': ['llvmlite>=0.36.0,<0.37.0'],
 'econml': ['econml'],
 'plotting': ['matplotlib>=3.5.3,<4.0.0'],
 'pydot': ['pydot>=1.4.2,<2.0.0'],
 'pygraphviz': ['pygraphviz>=1.9,<2.0']}

setup_kwargs = {
    'name': 'dowhy',
    'version': '0.9',
    'description': 'DoWhy is a Python library for causal inference that supports explicit modeling and testing of causal assumptions',
    'long_description': '|BuildStatus|_ |PyPiVersion|_ |PythonSupport|_ |Downloads|_\n\n.. |PyPiVersion| image:: https://img.shields.io/pypi/v/dowhy.svg\n.. _PyPiVersion: https://pypi.org/project/dowhy/\n\n.. |PythonSupport| image:: https://img.shields.io/pypi/pyversions/dowhy.svg\n.. _PythonSupport: https://pypi.org/project/dowhy/\n\n.. |BuildStatus| image:: https://github.com/microsoft/dowhy/workflows/Python%20package/badge.svg\n.. _BuildStatus: https://github.com/microsoft/dowhy/actions\n\n.. |Downloads| image:: https://pepy.tech/badge/dowhy\n.. _Downloads: https://pepy.tech/project/dowhy\n\n\n.. image:: dowhy-logo-large.png\n  :width: 50%\n  :align: center\n\n\\\n===============================\n\n  Introducing DoWhy and the 4 steps of causal inference | `Microsoft Research Blog <https://www.microsoft.com/en-us/research/blog/dowhy-a-library-for-causal-inference/>`_ | `Video Tutorial <https://note.microsoft.com/MSR-Webinar-DoWhy-Library-Registration-On-Demand.html>`_ | `Arxiv Paper <https://arxiv.org/abs/2011.04216>`_ | `Arxiv Paper (GCM-extension) <https://arxiv.org/abs/2206.06821>`_ | `Slides <https://www2.slideshare.net/AmitSharma315/dowhy-an-endtoend-library-for-causal-inference>`_\n\n  Read the `docs <https://py-why.github.io/dowhy/>`_ | Try it online! |Binder|_\n\n.. |Binder| image:: https://mybinder.org/badge_logo.svg\n.. _Binder: https://mybinder.org/v2/gh/microsoft/dowhy/main?filepath=docs%2Fsource%2F\n\n**Case Studies using DoWhy**: `Hotel booking cancellations <https://towardsdatascience.com/beyond-predictive-models-the-causal-story-behind-hotel-booking-cancellations-d29e8558cbaf>`_ | `Effect of customer loyalty programs <https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy_example_effect_of_memberrewards_program.ipynb>`_ | `Optimizing article headlines <https://medium.com/@akelleh/introducing-the-do-sampler-for-causal-inference-a3296ea9e78d>`_ | `Effect of home visits on infant health (IHDP) <https://towardsdatascience.com/implementing-causal-inference-a-key-step-towards-agi-de2cde8ea599>`_ | `Causes of customer churn/attrition <https://medium.com/geekculture/a-quickstart-for-causal-analysis-decision-making-with-dowhy-2ce2d4d1efa9>`_\n\n.. image:: https://raw.githubusercontent.com/microsoft/dowhy/main/docs/images/dowhy-schematic.png\n\nAs computing systems are more frequently and more actively intervening in societally critical domains such as healthcare, education, and governance, it is critical to correctly predict and understand the causal effects of these interventions. Without an A/B test, conventional machine learning methods, built on pattern recognition and correlational analyses, are insufficient for decision-making.\n\nMuch like machine learning libraries have done for prediction, **"DoWhy" is a Python library that aims to spark causal thinking and analysis**. DoWhy provides a principled four-step interface for causal inference that focuses on explicitly modeling causal assumptions and validating them as much as possible. The key feature of DoWhy is its state-of-the-art refutation API that can automatically test causal assumptions for any estimation method, thus making inference more robust and accessible to non-experts. DoWhy supports estimation of the average causal effect for backdoor, frontdoor, instrumental variable and other identification methods, and estimation of the conditional effect (CATE) through an integration with the EconML library.\n\nFor a quick introduction to causal inference, check out `amit-sharma/causal-inference-tutorial <https://github.com/amit-sharma/causal-inference-tutorial/>`_. We also gave a more comprehensive tutorial at the ACM Knowledge Discovery and Data Mining (`KDD 2018 <http://www.kdd.org/kdd2018/>`_) conference: `causalinference.gitlab.io/kdd-tutorial <http://causalinference.gitlab.io/kdd-tutorial/>`_. For an introduction to the four steps of causal inference and its implications for machine learning, you can access this video tutorial from Microsoft Research: `DoWhy Webinar <https://note.microsoft.com/MSR-Webinar-DoWhy-Library-Registration-On-Demand.html>`_.\n\nDocumentation for DoWhy is available at `py-why.github.io/dowhy <https://py-why.github.io/dowhy/>`_.\n\n.. i here comment toctree::\n.. i here comment   :maxdepth: 4\n.. i here comment   :caption: Contents:\n.. contents:: **Contents**\n\nNews\n-----\n**2022.05.27**:\n\n* **DoWhy now part of PyWhy**\n\n  We have moved DoWhy from microsoft/dowhy to py-why/dowhy. While GitHub will automatically\n  redirect your git command for cloning, pulling, etc., we recommend updating git remotes and bookmarks. Please note\n  that the **documentation** has now moved to https://py-why.github.io/dowhy with **no** redirect from the old URL.\n\n* **Experimental support for GCM-based inference**\n\n  We have started adding support for graphical causal model-based inference (or in short GCM-based). At the moment,\n  this includes support for interventions, counterfactuals, and attributing distribution changes. As part of this,\n  we also added features for Shapley value estimation and independence tests. We\'re still in the process of fleshing\n  everything out, including `documentation <https://py-why.github.io/dowhy/main/user_guide/gcm_based_inference/index.html>`_. Some of it is already on `main\n  <https://py-why.github.io/dowhy/main/user_guide/gcm_based_inference/index.html>`_, other parts are on feature branches (prefixed with ``gcm-``) with open\n  pull-requests, other parts will appear as new pull-requests in the next couple of weeks. Be sure to watch this space\n  here as we quickly expand functionality and documentation.\n\nThe need for causal inference\n----------------------------------\n\nPredictive models uncover patterns that connect the inputs and outcome in observed data. To intervene, however, we need to estimate the effect of changing an input from its current value, for which no data exists. Such questions, involving estimating a *counterfactual*, are common in decision-making scenarios.\n\n* Will it work?\n    * Does a proposed change to a system improve people\'s outcomes?\n* Why did it work?\n    * What led to a change in a system\'s outcome?\n* What should we do?\n    * What changes to a system are likely to improve outcomes for people?\n* What are the overall effects?\n    * How does the system interact with human behavior?\n    * What is the effect of a system\'s recommendations on people\'s activity?\n\nAnswering these questions requires causal reasoning. While many methods exist\nfor causal inference, it is hard to compare their assumptions and robustness of results. DoWhy makes three contributions,\n\n1. Provides a principled way of modeling a given problem as a causal graph so\n   that all assumptions are explicit.\n2. Provides a unified interface for many popular causal inference methods, combining the two major frameworks of graphical models and potential outcomes.\n3. Automatically tests for the validity of assumptions if possible and assesses\n   the robustness of the estimate to violations.\n\nTo see DoWhy in action, check out how it can be applied to estimate the effect\nof a subscription or rewards program for customers [`Rewards notebook\n<https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy_example_effect_of_memberrewards_program.ipynb>`_] and for implementing and evaluating causal inference methods on benchmark datasets like the `Infant Health and Development Program (IHDP) <https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy_ihdp_data_example.ipynb>`_ dataset, `Infant Mortality (Twins) <https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy_twins_example.ipynb>`_ dataset, and the `Lalonde Jobs <https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy_lalonde_example.ipynb>`_ dataset.\n\n\nInstallation\n-------------\n\nDoWhy support Python 3.8+. To install, you can use pip, poetry, or conda.\n\n**Latest Release**\n\nInstall the latest `release <https://pypi.org/project/dowhy/>`__ using pip.\n\n.. code:: shell\n\n   pip install dowhy\n\nInstall the latest `release <https://pypi.org/project/dowhy/>`__ using poetry.\n\n.. code:: shell\n\n   poetry add dowhy\n\nInstall the latest `release <https://anaconda.org/conda-forge/dowhy>`__ using conda.\n\n.. code:: shell\n\n   conda install -c conda-forge dowhy\n\nIf you face "Solving environment" problems with conda, then try :code:`conda update --all` and then install dowhy. If that does not work, then use :code:`conda config --set channel_priority false` and try to install again. If the problem persists, please `add your issue here <https://github.com/microsoft/dowhy/issues/197>`_.\n\n**Development Version**\n\nIf you prefer to use the latest dev version, your dependency management tool will need to point at our GitHub repository.\n\n.. code:: shell\n\n    pip install git+https://github.com/py-why/dowhy@main\n\n**Requirements**\n\nDoWhy requires a few dependencies. \nDetails on specific versions can be found in `pyproject.toml <./pyproject.toml>`_, under the `tool.poetry.dependencies` section.\n\nIf you face any problems, try installing dependencies manually.\n\n.. code:: shell\n\n    pip install \'<dependency-name>==<version>\'\n\nOptionally, if you wish to input graphs in the dot format, then install pydot (or pygraphviz).\n\n\nFor better-looking graphs, you can optionally install pygraphviz. To proceed,\nfirst install graphviz and then pygraphviz (on Ubuntu and Ubuntu WSL).\n\n.. code:: shell\n\n    sudo apt install graphviz libgraphviz-dev graphviz-dev pkg-config\n    ## from https://github.com/pygraphviz/pygraphviz/issues/71\n    pip install pygraphviz --install-option="--include-path=/usr/include/graphviz" \\\n    --install-option="--library-path=/usr/lib/graphviz/"\n\nSample causal inference analysis in DoWhy\n-------------------------------------------\nMost DoWhy\nanalyses for causal inference take 4 lines to write, assuming a\npandas dataframe df that contains the data:\n\n.. code:: python\n\n    from dowhy import CausalModel\n    import dowhy.datasets\n\n    # Load some sample data\n    data = dowhy.datasets.linear_dataset(\n        beta=10,\n        num_common_causes=5,\n        num_instruments=2,\n        num_samples=10000,\n        treatment_is_binary=True)\n\nDoWhy supports two formats for providing the causal graph: `gml <https://github.com/GunterMueller/UNI_PASSAU_FMI_Graph_Drawing>`_ (preferred) and `dot <http://www.graphviz.org/documentation/>`_. After loading in the data, we use the four main operations in DoWhy: *model*,\n*estimate*, *identify* and *refute*:\n\n.. code:: python\n\n    # I. Create a causal model from the data and given graph.\n    model = CausalModel(\n        data=data["df"],\n        treatment=data["treatment_name"],\n        outcome=data["outcome_name"],\n        graph=data["gml_graph"])\n\n    # II. Identify causal effect and return target estimands\n    identified_estimand = model.identify_effect()\n\n    # III. Estimate the target estimand using a statistical method.\n    estimate = model.estimate_effect(identified_estimand,\n                                     method_name="backdoor.propensity_score_matching")\n\n    # IV. Refute the obtained estimate using multiple robustness checks.\n    refute_results = model.refute_estimate(identified_estimand, estimate,\n                                           method_name="random_common_cause")\n\nDoWhy stresses on the interpretability of its output. At any point in the analysis,\nyou can inspect the untested assumptions, identified estimands (if any) and the\nestimate (if any). Here\'s a sample output of the linear regression estimator.\n\n.. image:: https://raw.githubusercontent.com/microsoft/dowhy/main/docs/images/regression_output.png\n\nFor a full code example, check out the `Getting Started with DoWhy <https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy_simple_example.ipynb>`_ notebook. You can also use Conditional Average Treatment Effect (CATE) estimation methods from other libraries such as EconML and CausalML, as shown in the `Conditional Treatment Effects <https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy-conditional-treatment-effects.ipynb>`_ notebook. For more examples of using DoWhy, check out the Jupyter notebooks in `docs/source/example_notebooks <https://github.com/microsoft/dowhy/tree/main/docs/source/example_notebooks/>`_ or try them online at `Binder <https://mybinder.org/v2/gh/microsoft/dowhy/main?filepath=docs%2Fsource%2F>`_.\n\n\nGCM-based inference (experimental)\n----------------------------------\n\nGraphical causal model-based inference, or GCM-based inference for short, is an experimental addition to DoWhy. For\ndetails, check out the `documentation for the gcm sub-package <https://py-why.github.io/dowhy/main/user_guide/gcm_based_inference/index.html>`_. The basic\nrecipe for this API works as follows:\n\n.. code:: python\n\n    # 1. Modeling cause-effect relationships as a structural causal model\n    #    (causal graph + functional causal models):\n    scm = gcm.StructuralCausalModel(nx.DiGraph([(\'X\', \'Y\'), (\'Y\', \'Z\')])) # X -> Y -> Z\n    scm.set_causal_mechanism(\'X\', gcm.EmpiricalDistribution())\n    scm.set_causal_mechanism(\'Y\', gcm.AdditiveNoiseModel(gcm.ml.create_linear_regressor()))\n    scm.set_causal_mechanism(\'Z\', gcm.AdditiveNoiseModel(gcm.ml.create_linear_regressor()))\n\n    # 2. Fitting the SCM to the data:\n    gcm.fit(scm, data)\n\n    # 3. Answering a causal query based on the SCM:\n    results = gcm.<causal_query>(scm, ...)\n\nWhere <causal_query> can be one of multiple functions explained in `Answering Causal Questions <https://py-why.github.io/dowhy/main/user_guide/gcm_based_inference/answering_causal_questions/index.html>`_.\n\n\nA high-level Pandas API\n-----------------------\n\nWe\'ve made an even simpler API for dowhy which is a light layer on top of the standard one. The goal is to make causal analysis much more like regular exploratory analysis. To use this API, simply\nimport :code:`dowhy.api`. This will magically add the :code:`causal` namespace to your\n:code:`pandas.DataFrame` s. Then,\nyou can use the namespace as follows.\n\n.. code:: python\n\n    import dowhy.api\n    import dowhy.datasets\n\n    data = dowhy.datasets.linear_dataset(beta=5,\n        num_common_causes=1,\n        num_instruments = 0,\n        num_samples=1000,\n        treatment_is_binary=True)\n\n    # data[\'df\'] is just a regular pandas.DataFrame\n    data[\'df\'].causal.do(x=\'v0\', # name of treatment variable\n                         variable_types={\'v0\': \'b\', \'y\': \'c\', \'W0\': \'c\'},\n                         outcome=\'y\',\n                         common_causes=[\'W0\']).groupby(\'v0\').mean().plot(y=\'y\', kind=\'bar\')\n\n.. image:: https://raw.githubusercontent.com/microsoft/dowhy/main/docs/images/do_barplot.png\n\nFor some methods, the :code:`variable_types` field must be specified. It should be a :code:`dict`, where the keys are\nvariable names, and values are \'o\' for ordered discrete, \'u\' for un-ordered discrete, \'d\' for discrete, or \'c\'\nfor continuous.\n\n**Note:If the** :code:`variable_types` **is not specified we make use of the following implicit conversions:**\n::\n\n   int -> \'c\'\n   float -> \'c\'\n   binary -> \'b\'\n   category -> \'d\'\n\n**Currently we have not added support for timestamps.**\n\nThe :code:`do` method in the causal namespace generates a random sample from $P(outcome|do(X=x))$ of the\nsame length as your data set, and returns this outcome as a new :code:`DataFrame`. You can continue to perform\nthe usual :code:`DataFrame` operations with this sample, and so you can compute statistics and create plots\nfor causal outcomes!\n\nThe :code:`do` method is built on top of the lower-level :code:`dowhy` objects, so can still take a graph and perform\nidentification automatically when you provide a graph instead of :code:`common_causes`.\n\nFor more details, check out the `Pandas API\n<https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy_causal_api.ipynb>`_ notebook or the `Do Sampler <https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/do_sampler_demo.ipynb>`_\nnotebook.\n\nGraphical Models and Potential Outcomes: Best of both worlds\n============================================================\nDoWhy builds on two of the most powerful frameworks for causal inference:\ngraphical models and potential outcomes. It uses graph-based criteria and\ndo-calculus for modeling assumptions and identifying a non-parametric causal effect.\nFor estimation, it switches to methods based primarily on potential outcomes.\n\nA unifying language for causal inference\n----------------------------------------\n\nDoWhy is based on a simple unifying language for causal inference. Causal\ninference may seem tricky, but almost all methods follow four key steps:\n\n1. Model a causal inference problem using assumptions.\n2. Identify an expression for the causal effect under these assumptions ("causal estimand").\n3. Estimate the expression using statistical methods such as matching or instrumental variables.\n4. Finally, verify the validity of the estimate using a variety of robustness checks.\n\nThis workflow can be captured by four key verbs in DoWhy:\n\n- model\n- identify\n- estimate\n- refute\n\nUsing these verbs, DoWhy implements a causal inference engine that can support\na variety of methods. *model* encodes prior knowledge as a formal causal graph, *identify* uses\ngraph-based methods to identify the causal effect, *estimate* uses\nstatistical methods for estimating the identified estimand, and finally *refute*\ntries to refute the obtained estimate by testing robustness to assumptions.\n\nKey differences compared to available causal inference software\n----------------------------------------------------------------\nDoWhy brings three key differences compared to available software for causal inference:\n\n**Explicit identifying assumptions**\n    Assumptions are first-class citizens in DoWhy.\n\n    Each analysis starts with a\n    building a causal model. The assumptions can be viewed graphically or in terms\n    of conditional independence statements. Wherever possible, DoWhy can also\n    automatically test for stated assumptions using observed data.\n\n**Separation between identification and estimation**\n    Identification is the causal problem. Estimation is simply a statistical problem.\n\n    DoWhy\n    respects this boundary and treats them separately. This focuses the causal\n    inference effort on identification, and frees up estimation using any\n    available statistical estimator for a target estimand. In addition, multiple\n    estimation methods can be used for a single identified_estimand and\n    vice-versa.\n\n**Automated robustness checks**\n    What happens when key identifying assumptions may not be satisfied?\n\n    The most critical, and often skipped, part of causal analysis is checking the\n    robustness of an estimate to unverified assumptions. DoWhy makes it easy to\n    automatically run sensitivity and robustness checks on the obtained estimate.\n\nFinally, DoWhy is easily extensible, allowing other implementations of the\nfour verbs to co-exist (e.g., we support implementations of the *estimation* verb from\nEconML and CausalML libraries). The four verbs are mutually independent, so their\nimplementations can be combined in any way.\n\n\n\nBelow are more details about the current implementation of each of these verbs.\n\nFour steps of causal inference\n===============================\n\nI. Model a causal problem\n-----------------------------\n\nDoWhy creates an underlying causal graphical model for each problem. This\nserves to make each causal assumption explicit. This graph need not be\ncomplete---you can provide a partial graph, representing prior\nknowledge about some of the variables. DoWhy automatically considers the rest\nof the variables as potential confounders.\n\nCurrently, DoWhy supports two formats for graph input: `gml <https://github.com/GunterMueller/UNI_PASSAU_FMI_Graph_Drawing>`_ (preferred) and\n`dot <http://www.graphviz.org/documentation/>`_. We strongly suggest to use gml as the input format, as it works well with networkx. You can provide the graph either as a .gml file or as a string. If you prefer to use dot format, you will need to install additional packages (pydot or pygraphviz, see the installation section above). Both .dot files and string format are supported.\n\nWhile not recommended, you can also specify common causes and/or instruments directly\ninstead of providing a graph.\n\nSupported formats for specifying causal assumptions\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n* **Graph**: Provide a causal graph in either gml or dot format. Can be a text file\n  or a string.\n* **Named variable sets**: Instead of the graph, provide variable names that\n  correspond to relevant categories, such as common causes, instrumental variables, effect\n  modifiers, frontdoor variables, etc.\n\nExamples of how to instantiate a causal model are in the `Getting Started\n<https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy_simple_example.ipynb>`_\nnotebook.\n\n.. i comment image:: causal_model.png\n\nII. Identify a target estimand under the model\n----------------------------------------------\n\nBased on the causal graph, DoWhy finds all possible ways of identifying a desired causal effect based on\nthe graphical model. It uses graph-based criteria and do-calculus to find\npotential ways find expressions that can identify the causal effect.\n\nSupported identification criteria\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n* Back-door criterion\n* Front-door criterion\n* Instrumental Variables\n* Mediation (Direct and indirect effect identification)\n\nDifferent notebooks illustrate how to use these identification criteria. Check\nout the `Simple Backdoor <https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy_confounder_example.ipynb>`_ notebook for the back-door criterion, and the `Simple IV <https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy-simple-iv-example.ipynb>`_ notebook for the instrumental variable criterion.\n\nIII. Estimate causal effect based on the identified estimand\n------------------------------------------------------------\n\nDoWhy supports methods based on both back-door criterion and instrumental\nvariables. It also provides a non-parametric confidence intervals and a permutation test for testing\nthe statistical significance of obtained estimate.\n\nSupported estimation methods\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n* Methods based on estimating the treatment assignment\n    * Propensity-based Stratification\n    * Propensity Score Matching\n    * Inverse Propensity Weighting\n\n* Methods based on estimating the outcome model\n    * Linear Regression\n    * Generalized Linear Models\n\n* Methods based on the instrumental variable equation\n    * Binary Instrument/Wald Estimator\n    * Two-stage least squares\n    * Regression discontinuity\n\n* Methods for front-door criterion and general mediation\n    * Two-stage linear regression\n\nExamples of using these methods are in the `Estimation methods\n<https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy_estimation_methods.ipynb>`_\nnotebook.\n\nUsing EconML and CausalML estimation methods in DoWhy\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\nIt is easy to call external estimation methods using DoWhy. Currently we\nsupport integrations with the `EconML <https://github.com/microsoft/econml>`_ and `CausalML <https://github.com/uber/causalml>`_ packages. Here\'s an example\nof estimating conditional treatment effects using EconML\'s double machine\nlearning estimator.\n\n.. code:: python\n\n\tfrom sklearn.preprocessing import PolynomialFeatures\n\tfrom sklearn.linear_model import LassoCV\n\tfrom sklearn.ensemble import GradientBoostingRegressor\n\tdml_estimate = model.estimate_effect(identified_estimand, method_name="backdoor.econml.dml.DML",\n                        control_value = 0,\n                        treatment_value = 1,\n                        target_units = lambda df: df["X0"]>1,\n                        confidence_intervals=False,\n                        method_params={\n                            "init_params":{\'model_y\':GradientBoostingRegressor(),\n                                           \'model_t\': GradientBoostingRegressor(),\n                                           \'model_final\':LassoCV(),\n                                           \'featurizer\':PolynomialFeatures(degree=1, include_bias=True)},\n                            "fit_params":{}}\n\t\t\t\t\t\t)\n\n\nMore examples are in the `Conditional Treatment Effects with DoWhy\n<https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy-conditional-treatment-effects.ipynb>`_ notebook.\n\nIV. Refute the obtained estimate\n-------------------------------------\nHaving access to multiple refutation methods to validate an effect estimate from a\ncausal estimator is\na key benefit of using DoWhy.\n\nSupported refutation methods\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n* **Add Random Common Cause**: Does the estimation method change its estimate after\n  we add an independent random variable as a common cause to the dataset?\n  (*Hint: It should not*)\n* **Placebo Treatment**: What happens to the estimated causal effect when we\n  replace the true treatment variable with an independent random variable?\n  (*Hint: the effect should go to zero*)\n* **Dummy Outcome**: What happens to the estimated causal effect when we replace\n  the true outcome variable with an independent random variable? (*Hint: The\n  effect should go to zero*)\n* **Simulated Outcome**: What happens to the estimated causal effect when we\n  replace the dataset with a simulated dataset based on a known data-generating\n  process closest to the given dataset? (*Hint: It should match the effect parameter\n  from the data-generating process*)\n* **Add Unobserved Common Causes**: How sensitive is the effect estimate when we\n  add an additional common cause (confounder) to the dataset that is correlated\n  with the treatment and the outcome? (*Hint: It should not be too sensitive*)\n* **Data Subsets Validation**: Does the estimated effect change significantly when\n  we replace the given dataset with a randomly selected subset? (*Hint: It\n  should not*)\n* **Bootstrap Validation**: Does the estimated effect change significantly when we\n  replace the given dataset with bootstrapped samples from the same dataset? (*Hint: It should not*)\n\nExamples of using refutation methods are in the `Refutations <https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy_refuter_notebook.ipynb>`_ notebook. For an advanced refutation that uses a simulated dataset based on user-provided or learnt data-generating processes, check out the `Dummy Outcome Refuter <https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy_demo_dummy_outcome_refuter.ipynb>`_ notebook.\nAs a practical example, `this notebook <https://github.com/microsoft/dowhy/blob/main/docs/source/example_notebooks/dowhy_refutation_testing.ipynb>`_ shows an application of refutation methods on evaluating effect estimators for the Infant Health and Development Program (IHDP) and Lalonde datasets.\n\nCiting this package\n====================\nIf you find DoWhy useful for your research work, please cite us as follows:\n\nAmit Sharma, Emre Kiciman, et al. DoWhy: A Python package for causal inference. 2019. https://github.com/microsoft/dowhy\n\nBibtex::\n\n  @misc{dowhy,\n  author={Sharma, Amit and Kiciman, Emre and others},\n  title={Do{W}hy: {A Python package for causal inference}},\n  howpublished={https://github.com/microsoft/dowhy},\n  year={2019}\n  }\n\nAlternatively, you can cite our Arxiv paper on DoWhy.\n\nAmit Sharma, Emre Kiciman. DoWhy: An End-to-End Library for Causal Inference. 2020. https://arxiv.org/abs/2011.04216\n\nBibtex::\n\n  @article{dowhypaper,\n  title={DoWhy: An End-to-End Library for Causal Inference},\n  author={Sharma, Amit and Kiciman, Emre},\n  journal={arXiv preprint arXiv:2011.04216},\n  year={2020}\n  }\n\nAnd if you find the gcm package useful for your work, please also cite us as:\n\nPatrick Blöbaum, Peter Götz, Kailash Budhathoki, Atalanti A. Mastakouri, Dominik Janzing. DoWhy-GCM: An extension of DoWhy for causal inference in graphical causal models. 2022. https://arxiv.org/abs/2206.06821\n\nBibtex::\n\n    @article{dowhy_gcm,\n      author = {Bl{\\"o}baum, Patrick and G{\\"o}tz, Peter and Budhathoki, Kailash and Mastakouri, Atalanti A. and Janzing, Dominik},\n      title = {DoWhy-GCM: An extension of DoWhy for causal inference in graphical causal models},\n      journal={arXiv preprint arXiv:2206.06821},\n      year={2022}\n    }\n\nRoadmap\n=======\nThe `projects <https://github.com/microsoft/dowhy/projects>`_ page lists the next steps for DoWhy. If you would like to contribute, have a look at the current projects. If you have a specific request for DoWhy, please `raise an issue <https://github.com/microsoft/dowhy/issues>`_.\n\nContributing\n============\n\nThis project welcomes contributions and suggestions. For a guide to contributing and a list of all contributors, check out `CONTRIBUTING.md <https://github.com/microsoft/dowhy/blob/main/CONTRIBUTING.md>`_ and our `docs for contributing code <https://github.com/py-why/dowhy/blob/main/docs/source/contributing/contributing-code.rst>`_. Our `contributor code of conduct is available here <https://github.com/py-why/governance/blob/main/CODE-OF-CONDUCT.md>`_. You can also join the DoWhy development channel on Discord: |discord|_\n\n.. |discord| image:: https://img.shields.io/discord/818456847551168542\n.. _discord: https://discord.gg/cSBGb3vsZb\n\n',
    'author': 'PyWhy Community',
    'author_email': 'amshar@microsoft.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/pywhy/dowhy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



import tempfile

import numpy as np
from flaky import flaky
from pytest import approx
from sklearn.model_selection import train_test_split

from dowhy.gcm.fcms import AdditiveNoiseModel, ClassifierFCM
from dowhy.gcm.ml.autolguon import AutoGluonClassifier, AutoGluonRegressor
from dowhy.gcm.util.general import shape_into_2d


@flaky(max_runs=3)
def test_given_linear_categorical_data_when_using_auto_gluon_classifier_linear_then_provides_good_fit():
    X, y = _generate_linear_classification_data(num_samples=1000, noise_std=0.2)

    x_train, x_test, y_train, y_test = train_test_split(X, y, test_size=0.5)

    y_test = shape_into_2d(y_test)
    with tempfile.TemporaryDirectory() as temp_dir:
        model = AutoGluonClassifier(path=temp_dir, time_limit=60, hyperparameters={"LR": {}})
        model.fit(x_train, y_train)

    assert np.sum(model.predict(x_test).squeeze() == y_test.squeeze()) / x_test.shape[0] == approx(1, abs=0.1)


@flaky(max_runs=3)
def test_given_linear_continuous_data_when_using_auto_gluon_regressor_linear_then_provides_good_fit():
    X, y, _ = _generate_linear_data(num_samples=2000, noise_std=0.2)

    x_train, x_test, y_train, y_test = train_test_split(X, y, test_size=0.5)

    y_test = shape_into_2d(y_test)
    with tempfile.TemporaryDirectory() as temp_dir:
        model = AutoGluonRegressor(path=temp_dir, time_limit=60, hyperparameters={"LR": {}})
        model.fit(x_train, y_train)

    assert np.std(model.predict(x_test) - y_test) == approx(0.2, abs=1)


def test_given_classifier_fcm_with_autogluon_classifier_model_when_drawing_samples_then_returns_samples():
    with tempfile.TemporaryDirectory() as temp_dir:
        autogluon_sem = ClassifierFCM(AutoGluonClassifier(path=temp_dir, time_limit=5))
        autogluon_sem.fit(np.random.normal(0, 1, (100, 3)), np.random.choice(3, 100, replace=True).astype(str))
    autogluon_sem.draw_samples(np.random.normal(0, 1, (5, 3)))


def test_given_anm_with_autogluon_regressor_model_when_drawing_samples_then_returns_samples():
    with tempfile.TemporaryDirectory() as temp_dir:
        autogluon_sem = AdditiveNoiseModel(AutoGluonClassifier(path=temp_dir, time_limit=5))
        autogluon_sem.fit(np.random.normal(0, 1, (100, 3)), np.random.normal(0, 1, (100, 1)))
    autogluon_sem.draw_samples(np.random.normal(0, 1, (5, 3)))


def _generate_linear_data(num_samples: int, noise_std: float):
    num_features = int(np.random.choice(20, 1) + 1)
    coefficients = np.random.uniform(-5, 5, num_features)

    X = np.random.normal(0, 1, (num_samples, num_features))
    y = np.sum(coefficients * X, axis=1) + np.random.normal(0, noise_std, num_samples)

    return X, y, coefficients


def _generate_linear_classification_data(num_samples: int, noise_std: float):
    X, y, _ = _generate_linear_data(num_samples, noise_std)

    y = y > np.median(y)

    return X, y.astype(str)



import numpy as np
from econml.grf._base_grf import BaseGRF
from econml.sklearn_extensions.model_selection import GridSearchCVList
from econml.utilities import cross_product
from scipy import stats
from sklearn.compose import ColumnTransformer, make_column_selector
from sklearn.ensemble import (
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.linear_model import LogisticRegression
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

from dowhy.utils.regression import create_polynomial_function, generate_moment_function, get_generic_regressor


def get_alpha_estimator(
    cv, X, max_degree=None, estimator_list=None, estimator_param_list=None, numeric_features=None, plugin_reisz=True
):
    """
    Finds the best estimator for reisz representer (alpha_s )

    :param cv: training and testing data indices obtained afteer Kfolding the dataset
    :param X: treatment+confounders
    :param max_degree: degree of the polynomial function used to approximate alpha_s
    :param param_grid_dict: python dictionary with parameters to tune the ReiszRepresenter estimator

    :returns: estimator for alpha_s

    This method assumes a binary T.
    """

    if plugin_reisz:
        if estimator_param_list is None:
            estimator_list = [
                RandomForestClassifier(n_estimators=100, random_state=120),
                Pipeline(
                    [
                        (
                            "scale",
                            ColumnTransformer([("num", StandardScaler(), numeric_features)], remainder="passthrough"),
                        ),
                        ("logistic_model", LogisticRegression()),
                    ]
                ),
                GradientBoostingClassifier(),
            ]
            estimator_param_list = [
                {"n_estimators": [50], "max_depth": [3, 4, 5], "min_samples_leaf": [10, 50]},
                {"logistic_model__C": [1, 0.01, 0.001, 1e-4, 1e-5]},
                {"learning_rate": [0.01, 0.001], "n_estimators": [50, 200]},
            ]
        W = X[:, 1:]  # assume 1-D treatment
        t = X[:, 0]
        propmodel = get_generic_regressor(
            cv=cv,
            X=W,
            Y=t,
            max_degree=max_degree,
            estimator_list=estimator_list,
            estimator_param_list=estimator_param_list,
            numeric_features=numeric_features,
        )
        reisz_fn = PluginReisz(propmodel)
    else:
        if estimator_param_list is None:
            estimator_param_list = [
                {
                    "reisz_functions": [create_polynomial_function(2), create_polynomial_function(1)],
                    "min_samples_leaf": [10, 50],
                    "min_var_fraction_leaf": [0.01, 0.1],
                    "l2_regularizer": [1e-5, 1e-3],
                    "max_depth": [5, None],
                }
            ]
        estimator = GridSearchCVList(
            [
                ReiszRepresenter(
                    reisz_functions=create_polynomial_function(max_degree),
                    moment_function=generate_moment_function,
                    min_samples_leaf=10,
                    min_var_fraction_leaf=0.01,
                    max_depth=5,
                )
            ],
            param_grid_list=estimator_param_list,
            scoring=reisz_scoring,
            cv=cv,
            verbose=0,
            n_jobs=-1,
        ).fit(X)
        reisz_fn = estimator.best_estimator_
    return reisz_fn


class PluginReisz:
    """Plugin reisz function for average treatment effect"""

    def __init__(self, propensity_model):
        self.propmodel = propensity_model

    def fit(self, X):
        W = X[:, 1:]  # assume 1-D treatment
        t = X[:, 0]
        self.propmodel.fit(W, t)
        return self

    def predict(self, X):
        W = X[:, 1:]  # assume 1-D treatment
        t = X[:, 0]
        preds = self.propmodel.predict_proba(W)
        weights = [1 / preds[i, t[i].astype(int)] for i in range(preds.shape[0])]
        weights = np.where(t == 0, -1, 1) * weights
        return weights

    def propensity(self, X):
        """
        P(T=1|W)
        """
        W = X[:, 1:]  # assume 1-D treatment
        return self.propmodel.predict_proba(W)[:, 1]


def reisz_scoring(g, X):
    """
    Returns loss for a estimator model
    :param g: model to be evaluated
    :param X: data for testing the model

    :returns: floating point number that quantifies the model prediction
    Loss(g) = E[2*m(X,g) - g(X)**2]
    """
    loss = np.mean(2 * generate_moment_function(W=X, g=g.predict) - g.predict(X) ** 2)
    return loss


class ReiszRepresenter(BaseGRF):
    """
    Generalized Random Forest to estimate Reisz Representer (RR)
    See: https://github.com/microsoft/EconML/blob/main/econml/grf/_base_grf.py
    :param reisz_functions: List of polynomial functions of n degree to approximate reisz representer created using create_polynomial_function
    :param moment_function: moment function m(W,g) whose expected value is used to calculate estimate
    :param l2_regularizer: l2 penalty while modeling (default = 1e-3)
    For tuning other parameters see https://econml.azurewebsites.net/_autosummary/econml.grf.CausalForest.html
    """

    def __init__(
        self,
        *,
        reisz_functions,
        moment_function=generate_moment_function,
        l2_regularizer=1e-3,
        n_estimators=100,
        criterion="mse",
        max_depth=None,
        min_samples_split=10,
        min_samples_leaf=5,
        min_weight_fraction_leaf=0.0,
        min_var_fraction_leaf=None,
        min_var_leaf_on_val=False,
        max_features="auto",
        min_impurity_decrease=0.0,
        max_samples=0.45,
        min_balancedness_tol=0.45,
        honest=True,
        inference=True,
        fit_intercept=True,
        subforest_size=4,
        n_jobs=-1,
        random_state=None,
        verbose=0,
        warm_start=False,
    ):

        self.reisz_functions = reisz_functions
        self.moment_function = moment_function
        self.l2_regularizer = l2_regularizer
        self.num_reisz_functions = len(self.reisz_functions)

        super().__init__(
            n_estimators=n_estimators,
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            min_var_fraction_leaf=min_var_fraction_leaf,
            min_var_leaf_on_val=min_var_leaf_on_val,
            max_features=max_features,
            min_impurity_decrease=min_impurity_decrease,
            max_samples=max_samples,
            min_balancedness_tol=min_balancedness_tol,
            honest=honest,
            inference=inference,
            fit_intercept=False,
            subforest_size=subforest_size,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
        )

    def _get_alpha_and_pointJ(self, X, T, y):
        n_riesz_feats = len(self.reisz_functions)
        TX = np.hstack([T, X])
        riesz_feats = np.hstack([feat_fn(TX) for feat_fn in self.reisz_functions])
        mfeats = np.hstack([self.moment_function(TX, feat_fn) for feat_fn in self.reisz_functions])
        alpha = np.zeros((X.shape[0], n_riesz_feats))
        alpha[:, :n_riesz_feats] = mfeats
        riesz_cov_matrix = cross_product(riesz_feats, riesz_feats).reshape((X.shape[0], n_riesz_feats, n_riesz_feats))
        penalty = self.l2_regularizer * np.eye(n_riesz_feats)
        penalty[0, 0] = 0
        pointJ = riesz_cov_matrix + penalty
        return alpha, pointJ.reshape((X.shape[0], -1))

    def _get_n_outputs_decomposition(self, X, T, y):
        n_relevant_outputs = len(self.reisz_functions)
        n_outputs = n_relevant_outputs
        return n_outputs, n_relevant_outputs

    def _translate(self, point, TX):
        riesz_feats = np.hstack([feat_fn(TX) for feat_fn in self.reisz_functions])
        n_riesz_feats = riesz_feats.shape[1]
        riesz = np.sum(point[:, :n_riesz_feats] * riesz_feats, axis=1)
        return riesz

    def fit(self, X):
        return super().fit(X[:, 1:], X[:, [0]], X[:, [0]])

    def predict(self, X_test):
        point = super().predict(X_test[:, 1:])
        return self._translate(point, X_test)



from typing import cast

import networkx as nx
import numpy as np
import pandas as pd
import pytest
from flaky import flaky
from pytest import approx
from scipy import stats
from scipy.stats import norm

from dowhy.gcm import (
    AdditiveNoiseModel,
    ClassifierFCM,
    EmpiricalDistribution,
    PostNonlinearModel,
    ProbabilisticCausalModel,
    ScipyDistribution,
    StructuralCausalModel,
    draw_samples,
    fit,
)
from dowhy.gcm.divergence import estimate_kl_divergence_continuous
from dowhy.gcm.ml import (
    SklearnRegressionModel,
    create_linear_regressor,
    create_linear_regressor_with_given_parameters,
    create_logistic_regression_classifier,
)
from dowhy.gcm.ml.regression import (
    InvertibleExponentialFunction,
    InvertibleIdentityFunction,
    InvertibleLogarithmicFunction,
)


def test_given_linear_data_when_fit_causal_graph_with_linear_anm_then_learns_correct_coefficients():
    scm = ProbabilisticCausalModel(nx.DiGraph([("X0", "X1")]))
    scm.set_causal_mechanism("X0", ScipyDistribution(stats.norm, loc=0, scale=1))
    scm.set_causal_mechanism("X1", AdditiveNoiseModel(prediction_model=create_linear_regressor()))

    X0 = scm.causal_mechanism("X0").draw_samples(1000).squeeze()
    test_data = pd.DataFrame({"X0": X0, "X1": X0 * 2 + 2 + np.random.normal(0, 0.1, 1000)})
    fit(scm, test_data)

    assert scm.causal_mechanism("X1").prediction_model.sklearn_model.coef_ == approx(np.array([2]), abs=0.02)
    assert scm.causal_mechanism("X1").prediction_model.sklearn_model.intercept_ == approx(2, abs=0.02)


def test_given_linear_data_when_draw_samples_from_fitted_anm_then_generates_correct_marginal_distribution():
    scm = ProbabilisticCausalModel(nx.DiGraph([("X0", "X1")]))
    scm.set_causal_mechanism("X0", ScipyDistribution(stats.norm, loc=0, scale=1))
    scm.set_causal_mechanism("X1", AdditiveNoiseModel(prediction_model=create_linear_regressor()))

    X0 = scm.causal_mechanism("X0").draw_samples(1000).squeeze()
    test_data = pd.DataFrame({"X0": X0, "X1": X0 * 2 + 2 + np.random.normal(0, 0.1, 1000)})
    fit(scm, test_data)

    generated_samples = scm.causal_mechanism("X1").draw_samples(np.array([1] * 1000))
    assert np.mean(generated_samples) == approx(4, abs=0.05)
    assert np.std(generated_samples) == approx(0.1, abs=0.05)

    generated_samples = scm.causal_mechanism("X1").draw_samples(np.array([2] * 1000))
    assert np.mean(generated_samples) == approx(6, abs=0.05)
    assert np.std(generated_samples) == approx(0.1, abs=0.05)
    assert estimate_kl_divergence_continuous(
        test_data["X1"].to_numpy(), draw_samples(scm, 1000)["X1"].to_numpy()
    ) == approx(0, abs=0.05)


@flaky(max_runs=5)
def test_given_categorical_input_data_when_fit_causal_graph_with_linear_anm_then_learns_correct_coefficients():
    scm = StructuralCausalModel(nx.DiGraph([("X0", "X2"), ("X1", "X2")]))
    scm.set_causal_mechanism("X0", ScipyDistribution(stats.norm, loc=0, scale=1))
    scm.set_causal_mechanism("X1", EmpiricalDistribution())
    scm.set_causal_mechanism("X2", AdditiveNoiseModel(prediction_model=create_linear_regressor()))

    training_data = _generate_data_with_categorical_input()
    fit(scm, data=training_data)

    # Mean from the categorical part is: (-5 + 5+ 10) / 3 = 10/3
    assert scm.causal_mechanism("X2").prediction_model.sklearn_model.coef_ == approx(
        np.array([2, -5 - 10 / 3, 10 - 10 / 3, 5 - 10 / 3]), abs=0.02
    )
    assert scm.causal_mechanism("X2").prediction_model.sklearn_model.intercept_ == approx(10 / 3, abs=0.02)


def test_given_categorical_input_data_when_draw_from_fitted_causal_graph_with_linear_anm_then_generates_correct_marginal_distribution():
    scm = StructuralCausalModel(nx.DiGraph([("X0", "X2"), ("X1", "X2")]))
    scm.set_causal_mechanism("X0", ScipyDistribution(stats.norm, loc=0, scale=1))
    scm.set_causal_mechanism("X1", EmpiricalDistribution())
    scm.set_causal_mechanism("X2", AdditiveNoiseModel(prediction_model=create_linear_regressor()))

    training_data = _generate_data_with_categorical_input()
    fit(scm, data=training_data)

    assert scm.causal_mechanism("X2").evaluate(np.array([[2, "1"]], dtype=object), np.array([0])) == approx(14)

    test_data = training_data.to_numpy()
    assert scm.causal_mechanism("X2").evaluate(test_data[:, :2], np.array([0] * 1)) == approx(
        test_data[:, 2].astype(float).reshape(-1, 1)
    )

    assert estimate_kl_divergence_continuous(test_data[:, 2], draw_samples(scm, 1000)["X2"].to_numpy()) == approx(
        0, abs=0.05
    )


def test_given_non_string_data_when_try_to_fit_classifier_fcm_then_throws_error():
    scm = StructuralCausalModel(nx.DiGraph([("X0", "X1")]))
    scm.set_causal_mechanism("X0", ScipyDistribution(stats.norm, loc=0, scale=1))
    scm.set_causal_mechanism("X1", ClassifierFCM(classifier_model=create_logistic_regression_classifier()))

    X0 = np.random.normal(0, 1, 1000)
    X1 = (X0 > np.median(X0)).astype(int)

    with pytest.raises(ValueError):
        fit(scm, data=(pd.DataFrame({"X0": X0, "X1": X1})))


def test_when_draw_from_classifier_fcm_then_returns_string_samples():
    scm = StructuralCausalModel(nx.DiGraph([("X0", "X1")]))
    scm.set_causal_mechanism("X0", ScipyDistribution(stats.norm, loc=0, scale=1))
    scm.set_causal_mechanism("X1", ClassifierFCM(classifier_model=create_logistic_regression_classifier()))

    X0 = np.random.normal(0, 1, 1000)
    X1 = (X0 > np.median(X0)).astype(str)

    training_data = pd.DataFrame({"X0": X0, "X1": X1})

    fit(scm, training_data)

    for val in scm.causal_mechanism("X1").draw_samples(X0):
        assert isinstance(val[0], str)


@flaky(max_runs=5)
def test_when_fit_classifier_fcm_with_categorical_inputs_then_returns_expected_results():
    scm = StructuralCausalModel(nx.DiGraph([("X0", "X2"), ("X1", "X2")]))
    scm.set_causal_mechanism("X0", ScipyDistribution(stats.norm, loc=0, scale=1))
    scm.set_causal_mechanism("X1", EmpiricalDistribution())
    scm.set_causal_mechanism("X2", ClassifierFCM(classifier_model=create_logistic_regression_classifier()))

    training_data = _generate_data_with_categorical_input()
    X2 = training_data["X2"].to_numpy()
    training_data["X2"] = (X2 > np.median(X2)).astype(str)
    X2 = training_data["X2"].to_numpy()
    fit(scm, training_data)

    x2_fcm = cast(ClassifierFCM, scm.causal_mechanism("X2"))
    assert x2_fcm.estimate_probabilities(np.array([[2, "1"]], dtype=object)) == approx(np.array([[0, 1]]), abs=0.01)

    test_data = training_data.to_numpy()
    X2[X2 == "True"] = 1
    X2[X2 == "False"] = 0
    assert np.sum(np.argmax(x2_fcm.estimate_probabilities(test_data[:, :2]), axis=1) != X2.astype(int)) < 20

    _, counts = np.unique(x2_fcm.draw_samples(test_data[:, :2]), return_counts=True)
    assert counts / 1000 == approx(np.array([0.5, 0.5]), abs=0.05)


def test_when_clone_additive_noise_models_with_scipy_distribution_then_clone_has_correct_models():
    org_model = AdditiveNoiseModel(create_linear_regressor(), ScipyDistribution(norm))
    clone_1 = org_model.clone()
    clone_2 = org_model.clone()

    org_model.fit(np.array([1, 2, 3, 4, 5]), np.array([1, 2, 3, 4, 5]))
    clone_1.fit(np.array([1, 2, 3, 4, 5]), np.array([1, 2, 3, 4, 5]))
    clone_2.fit(np.array([1, 2, 3, 4, 5]), np.array([1, 2, 3, 4, 5]))

    assert isinstance(clone_1, AdditiveNoiseModel)
    assert isinstance(clone_1.prediction_model, SklearnRegressionModel)
    assert isinstance(clone_1.noise_model, ScipyDistribution)
    assert isinstance(clone_2, AdditiveNoiseModel)
    assert isinstance(clone_2.prediction_model, SklearnRegressionModel)
    assert isinstance(clone_2.noise_model, ScipyDistribution)


def test_given_simple_linear_data_when_fit_post_non_linear_sem_with_invertible_identity_then_returns_expected_results():
    X = np.random.normal(0, 1, 1000)
    N = np.random.normal(0, 0.1, 1000)
    Y = 2 * X + N

    sem_ground_truth = PostNonlinearModel(
        create_linear_regressor_with_given_parameters(coefficients=np.array([2])),
        EmpiricalDistribution(),
        InvertibleIdentityFunction(),
    )

    assert sem_ground_truth.estimate_noise(Y, X).reshape(-1) == approx(N)
    assert sem_ground_truth.evaluate(np.array([2]), np.array([0])).squeeze() == 4
    assert sem_ground_truth.evaluate(np.array([2]), np.array([1])).squeeze() == 4 + 1

    sem_fitted = PostNonlinearModel(create_linear_regressor(), EmpiricalDistribution(), InvertibleIdentityFunction())
    sem_fitted.fit(X, Y)

    assert sem_fitted.prediction_model.sklearn_model.coef_ == approx(np.array([2]), abs=0.05)
    assert np.mean(sem_fitted.draw_samples(np.array([2] * 1000))) == approx(4, abs=0.05)


def test_given_exponential_data_when_fit_post_non_linear_sem_with_invertible_exponential_function_then_returns_expected_results():
    X = np.random.normal(0, 1, 1000)
    N = np.random.normal(0, 0.1, 1000)
    Y = np.exp(2 * X + N)

    sem_ground_truth = PostNonlinearModel(
        create_linear_regressor_with_given_parameters(coefficients=np.array([2])),
        EmpiricalDistribution(),
        InvertibleExponentialFunction(),
    )

    assert sem_ground_truth.estimate_noise(Y, X).reshape(-1) == approx(N)
    assert sem_ground_truth.evaluate(np.array([2]), np.array([0])).squeeze() == np.exp(4)
    assert sem_ground_truth.evaluate(np.array([2]), np.array([1])).squeeze() == np.exp(4 + 1)

    sem_fitted = PostNonlinearModel(create_linear_regressor(), EmpiricalDistribution(), InvertibleExponentialFunction())
    sem_fitted.fit(X, Y)

    assert sem_fitted.prediction_model.sklearn_model.coef_ == approx(np.array([2]), abs=0.05)


def test_given_logarithmic_data_when_fit_post_non_linear_sem_with_invertible_logarithmic_function_then_returns_expected_results():
    X = abs(np.random.normal(0, 1, 1000))
    N = abs(np.random.normal(0, 0.1, 1000))
    Y = np.log(2 * X + N)

    sem_ground_truth = PostNonlinearModel(
        create_linear_regressor_with_given_parameters(coefficients=np.array([2])),
        EmpiricalDistribution(),
        InvertibleLogarithmicFunction(),
    )

    assert sem_ground_truth.estimate_noise(Y, X).reshape(-1) == approx(N)
    assert sem_ground_truth.evaluate(np.array([2]), np.array([0])).squeeze() == np.log(4)
    assert sem_ground_truth.evaluate(np.array([2]), np.array([1])).squeeze() == np.log(4 + 1)

    sem_fitted = PostNonlinearModel(create_linear_regressor(), EmpiricalDistribution(), InvertibleLogarithmicFunction())
    sem_fitted.fit(X, Y)

    assert sem_fitted.prediction_model.sklearn_model.coef_ == approx(np.array([2]), abs=0.05)


def _generate_data_with_categorical_input():
    X0 = np.random.normal(0, 1, 1000)
    X1 = np.random.choice(3, 1000).astype(str)
    X2 = []

    for i in range(1000):
        tmp_value = 2 * X0[i]

        if X1[i] == "0":
            tmp_value -= 5
        elif X1[i] == "1":
            tmp_value += 10
        else:
            tmp_value += 5

        X2.append(tmp_value)

    return pd.DataFrame({"X0": X0, "X1": X1, "X2": X2})



import itertools
import logging
from enum import Enum
from typing import Dict, List, Optional, Union

import sympy as sp
import sympy.stats as spstats

from dowhy.causal_graph import CausalGraph
from dowhy.causal_identifier.efficient_backdoor import EfficientBackdoor
from dowhy.causal_identifier.identified_estimand import IdentifiedEstimand
from dowhy.utils.api import parse_state

logger = logging.getLogger(__name__)


class EstimandType(Enum):
    # Average total effect
    NONPARAMETRIC_ATE = "nonparametric-ate"
    # Natural direct effect
    NONPARAMETRIC_NDE = "nonparametric-nde"
    # Natural indirect effect
    NONPARAMETRIC_NIE = "nonparametric-nie"
    # Controlled direct effect
    NONPARAMETRIC_CDE = "nonparametric-cde"


class BackdoorAdjustment(Enum):

    # Backdoor method names
    BACKDOOR_DEFAULT = "default"
    BACKDOOR_EXHAUSTIVE = "exhaustive-search"
    BACKDOOR_MIN = "minimal-adjustment"
    BACKDOOR_MAX = "maximal-adjustment"
    BACKDOOR_EFFICIENT = "efficient-adjustment"
    BACKDOOR_MIN_EFFICIENT = "efficient-minimal-adjustment"
    BACKDOOR_MINCOST_EFFICIENT = "efficient-mincost-adjustment"


MAX_BACKDOOR_ITERATIONS = 100000

METHOD_NAMES = {
    BackdoorAdjustment.BACKDOOR_DEFAULT,
    BackdoorAdjustment.BACKDOOR_EXHAUSTIVE,
    BackdoorAdjustment.BACKDOOR_MIN,
    BackdoorAdjustment.BACKDOOR_MAX,
    BackdoorAdjustment.BACKDOOR_EFFICIENT,
    BackdoorAdjustment.BACKDOOR_MIN_EFFICIENT,
    BackdoorAdjustment.BACKDOOR_MINCOST_EFFICIENT,
}
EFFICIENT_METHODS = {
    BackdoorAdjustment.BACKDOOR_EFFICIENT,
    BackdoorAdjustment.BACKDOOR_MIN_EFFICIENT,
    BackdoorAdjustment.BACKDOOR_MINCOST_EFFICIENT,
}
DEFAULT_BACKDOOR_METHOD = BackdoorAdjustment.BACKDOOR_DEFAULT


class AutoIdentifier:
    """Class that implements different identification methods.

    Currently supports backdoor and instrumental variable identification methods. The identification is based on the causal graph provided.

    This class is for backwards compatibility with CausalModel
    Will be deprecated in the future in favor of function call auto_identify_effect()

    """

    def __init__(
        self,
        estimand_type: EstimandType,
        backdoor_adjustment: BackdoorAdjustment = BackdoorAdjustment.BACKDOOR_DEFAULT,
        proceed_when_unidentifiable: bool = False,
        optimize_backdoor: bool = False,
        costs: Optional[List] = None,
    ):
        self.estimand_type = estimand_type
        self.backdoor_adjustment = backdoor_adjustment
        self._proceed_when_unidentifiable = proceed_when_unidentifiable
        self.optimize_backdoor = optimize_backdoor
        self.costs = costs
        self.logger = logging.getLogger(__name__)

    def identify_effect(
        self,
        graph: CausalGraph,
        treatment_name: Union[str, List[str]],
        outcome_name: Union[str, List[str]],
        conditional_node_names: List[str] = None,
        **kwargs,
    ):
        estimand = identify_effect_auto(
            graph,
            treatment_name,
            outcome_name,
            self.estimand_type,
            conditional_node_names,
            self.backdoor_adjustment,
            self._proceed_when_unidentifiable,
            self.optimize_backdoor,
            self.costs,
            **kwargs,
        )

        estimand.identifier = self

        return estimand

    def identify_backdoor(
        self,
        graph: CausalGraph,
        treatment_name: List[str],
        outcome_name: str,
        include_unobserved: bool = False,
        dseparation_algo: str = "default",
        direct_effect: bool = False,
    ):
        return identify_backdoor(
            graph,
            treatment_name,
            outcome_name,
            self.backdoor_adjustment,
            include_unobserved,
            dseparation_algo,
            direct_effect,
        )


def identify_effect_auto(
    graph: CausalGraph,
    treatment_name: Union[str, List[str]],
    outcome_name: Union[str, List[str]],
    estimand_type: EstimandType,
    conditional_node_names: List[str] = None,
    backdoor_adjustment: BackdoorAdjustment = BackdoorAdjustment.BACKDOOR_DEFAULT,
    proceed_when_unidentifiable: bool = False,
    optimize_backdoor: bool = False,
    costs: Optional[List] = None,
    **kwargs,
) -> IdentifiedEstimand:
    """Main method that returns an identified estimand (if one exists).

    If estimand_type is non-parametric ATE, then  uses backdoor, instrumental variable and frontdoor identification methods,  to check if an identified estimand exists, based on the causal graph.

    :param optimize_backdoor: if True, uses an optimised algorithm to compute the backdoor sets
    :param costs: non-negative costs associated with variables in the graph. Only used
    for estimand_type='non-parametric-ate' and backdoor_adjustment='efficient-mincost-adjustment'. If
    no costs are provided by the user, and backdoor_adjustment='efficient-mincost-adjustment', costs
    are assumed to be equal to one for all variables in the graph.
    :param conditional_node_names: variables that are used to determine treatment. If none are
    provided, it is assumed that the intervention is static.
    :returns:  target estimand, an instance of the IdentifiedEstimand class
    """

    treatment_name = parse_state(treatment_name)
    outcome_name = parse_state(outcome_name)

    # First, check if there is a directed path from action to outcome
    if not graph.has_directed_path(treatment_name, outcome_name):
        logger.warn("No directed path from treatment to outcome. Causal Effect is zero.")
        return IdentifiedEstimand(
            None,
            treatment_variable=treatment_name,
            outcome_variable=outcome_name,
            no_directed_path=True,
        )
    if estimand_type == EstimandType.NONPARAMETRIC_ATE:
        return identify_ate_effect(
            graph,
            treatment_name,
            outcome_name,
            backdoor_adjustment,
            optimize_backdoor,
            estimand_type,
            costs,
            conditional_node_names,
            proceed_when_unidentifiable,
        )
    elif estimand_type == EstimandType.NONPARAMETRIC_NDE:
        return identify_nde_effect(
            graph, treatment_name, outcome_name, backdoor_adjustment, estimand_type, proceed_when_unidentifiable
        )
    elif estimand_type == EstimandType.NONPARAMETRIC_NIE:
        return identify_nie_effect(
            graph, treatment_name, outcome_name, backdoor_adjustment, estimand_type, proceed_when_unidentifiable
        )
    elif estimand_type == EstimandType.NONPARAMETRIC_CDE:
        return identify_cde_effect(
            graph, treatment_name, outcome_name, backdoor_adjustment, estimand_type, proceed_when_unidentifiable
        )
    else:
        raise ValueError(
            "Estimand type is not supported. Use either {0}, {1}, or {2}.".format(
                EstimandType.NONPARAMETRIC_ATE,
                EstimandType.NONPARAMETRIC_CDE,
                EstimandType.NONPARAMETRIC_NDE,
                EstimandType.NONPARAMETRIC_NIE,
            )
        )


def identify_ate_effect(
    graph: CausalGraph,
    treatment_name: List[str],
    outcome_name: str,
    backdoor_adjustment: BackdoorAdjustment,
    optimize_backdoor: bool,
    estimand_type: EstimandType,
    costs: List,
    conditional_node_names: List[str] = None,
    proceed_when_unidentifiable: bool = False,
):
    estimands_dict = {}
    mediation_first_stage_confounders = None
    mediation_second_stage_confounders = None
    ### 1. BACKDOOR IDENTIFICATION
    # Pick algorithm to compute backdoor sets according to method chosen
    if backdoor_adjustment not in EFFICIENT_METHODS:
        # First, checking if there are any valid backdoor adjustment sets
        if optimize_backdoor == False:
            backdoor_sets = identify_backdoor(graph, treatment_name, outcome_name, backdoor_adjustment)
        else:
            from dowhy.causal_identifier.backdoor import Backdoor

            path = Backdoor(graph._graph, treatment_name, outcome_name)
            backdoor_sets = path.get_backdoor_vars()
    elif backdoor_adjustment in EFFICIENT_METHODS:
        backdoor_sets = identify_efficient_backdoor(
            graph, backdoor_adjustment, costs, conditional_node_names=conditional_node_names
        )
    estimands_dict, backdoor_variables_dict = build_backdoor_estimands_dict(
        graph, treatment_name, outcome_name, backdoor_sets, estimands_dict
    )
    # Setting default "backdoor" identification adjustment set
    default_backdoor_id = get_default_backdoor_set_id(graph, treatment_name, outcome_name, backdoor_variables_dict)
    if len(backdoor_variables_dict) > 0:
        estimands_dict["backdoor"] = estimands_dict.get(str(default_backdoor_id), None)
        backdoor_variables_dict["backdoor"] = backdoor_variables_dict.get(str(default_backdoor_id), None)
    else:
        estimands_dict["backdoor"] = None
    ### 2. INSTRUMENTAL VARIABLE IDENTIFICATION
    # Now checking if there is also a valid iv estimand
    instrument_names = graph.get_instruments(treatment_name, outcome_name)
    logger.info("Instrumental variables for treatment and outcome:" + str(instrument_names))
    if len(instrument_names) > 0:
        iv_estimand_expr = construct_iv_estimand(
            treatment_name,
            outcome_name,
            instrument_names,
        )
        logger.debug("Identified expression = " + str(iv_estimand_expr))
        estimands_dict["iv"] = iv_estimand_expr
    else:
        estimands_dict["iv"] = None

    ### 3. FRONTDOOR IDENTIFICATION
    # Now checking if there is a valid frontdoor variable
    frontdoor_variables_names = identify_frontdoor(graph, treatment_name, outcome_name)
    logger.info("Frontdoor variables for treatment and outcome:" + str(frontdoor_variables_names))
    if len(frontdoor_variables_names) > 0:
        frontdoor_estimand_expr = construct_frontdoor_estimand(
            treatment_name,
            outcome_name,
            frontdoor_variables_names,
        )
        logger.debug("Identified expression = " + str(frontdoor_estimand_expr))
        estimands_dict["frontdoor"] = frontdoor_estimand_expr
        mediation_first_stage_confounders = identify_mediation_first_stage_confounders(
            graph, treatment_name, outcome_name, frontdoor_variables_names, backdoor_adjustment
        )
        mediation_second_stage_confounders = identify_mediation_second_stage_confounders(
            graph, treatment_name, frontdoor_variables_names, outcome_name, backdoor_adjustment
        )
    else:
        estimands_dict["frontdoor"] = None

    # Finally returning the estimand object
    estimand = IdentifiedEstimand(
        None,
        treatment_variable=treatment_name,
        outcome_variable=outcome_name,
        estimand_type=estimand_type,
        estimands=estimands_dict,
        backdoor_variables=backdoor_variables_dict,
        instrumental_variables=instrument_names,
        frontdoor_variables=frontdoor_variables_names,
        mediation_first_stage_confounders=mediation_first_stage_confounders,
        mediation_second_stage_confounders=mediation_second_stage_confounders,
        default_backdoor_id=default_backdoor_id,
    )
    return estimand


def identify_cde_effect(
    graph: CausalGraph,
    treatment_name: List[str],
    outcome_name: str,
    backdoor_adjustment: BackdoorAdjustment,
    estimand_type: EstimandType,
    proceed_when_unidentifiable: bool = False,
):
    """Identify controlled direct effect. For a definition, see Vanderwheele (2011).
    Controlled direct and mediated effects: definition, identification and bounds.
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4193506/

    Using do-calculus rules, identification yields a adjustment set.
    It is based on the principle that under a graph where the direct edge from treatment
    to outcome is removed, conditioning on the adjustment set should d-separate
    treatment and outcome.
    """
    estimands_dict = {}
    # Pick algorithm to compute backdoor sets according to method chosen
    backdoor_sets = identify_backdoor(graph, treatment_name, outcome_name, backdoor_adjustment, direct_effect=True)
    estimands_dict, backdoor_variables_dict = build_backdoor_estimands_dict(
        graph, treatment_name, outcome_name, backdoor_sets, estimands_dict
    )
    # Setting default "backdoor" identification adjustment set
    default_backdoor_id = get_default_backdoor_set_id(graph, treatment_name, outcome_name, backdoor_variables_dict)
    if len(backdoor_variables_dict) > 0:
        estimands_dict["backdoor"] = estimands_dict.get(str(default_backdoor_id), None)
        backdoor_variables_dict["backdoor"] = backdoor_variables_dict.get(str(default_backdoor_id), None)
    else:
        estimands_dict["backdoor"] = None

    # Finally returning the estimand object
    estimand = IdentifiedEstimand(
        None,
        treatment_variable=treatment_name,
        outcome_variable=outcome_name,
        estimand_type=estimand_type,
        estimands=estimands_dict,
        backdoor_variables=backdoor_variables_dict,
        instrumental_variables=None,
        frontdoor_variables=None,
        mediation_first_stage_confounders=None,
        mediation_second_stage_confounders=None,
        default_backdoor_id=default_backdoor_id,
    )
    return estimand


def identify_nie_effect(
    graph: CausalGraph,
    treatment_name: List[str],
    outcome_name: str,
    backdoor_adjustment: BackdoorAdjustment,
    estimand_type: EstimandType,
    proceed_when_unidentifiable: bool = False,
):
    estimands_dict = {}
    ### 1. FIRST DOING BACKDOOR IDENTIFICATION
    # First, checking if there are any valid backdoor adjustment sets
    backdoor_sets = identify_backdoor(graph, treatment_name, outcome_name, backdoor_adjustment)
    estimands_dict, backdoor_variables_dict = build_backdoor_estimands_dict(
        graph, treatment_name, outcome_name, backdoor_sets, estimands_dict
    )
    # Setting default "backdoor" identification adjustment set
    default_backdoor_id = get_default_backdoor_set_id(graph, treatment_name, outcome_name, backdoor_variables_dict)
    backdoor_variables_dict["backdoor"] = backdoor_variables_dict.get(str(default_backdoor_id), None)

    ### 2. SECOND, CHECKING FOR MEDIATORS
    # Now checking if there are valid mediator variables
    estimands_dict = {}  # Need to reinitialize this dictionary to avoid including the backdoor sets
    mediation_first_stage_confounders = None
    mediation_second_stage_confounders = None
    mediators_names = identify_mediation(graph, treatment_name, outcome_name)
    logger.info("Mediators for treatment and outcome:" + str(mediators_names))
    if len(mediators_names) > 0:
        mediation_estimand_expr = construct_mediation_estimand(
            estimand_type,
            treatment_name,
            outcome_name,
            mediators_names,
        )
        logger.debug("Identified expression = " + str(mediation_estimand_expr))
        estimands_dict["mediation"] = mediation_estimand_expr
        mediation_first_stage_confounders = identify_mediation_first_stage_confounders(
            graph, treatment_name, outcome_name, mediators_names, backdoor_adjustment
        )
        mediation_second_stage_confounders = identify_mediation_second_stage_confounders(
            graph, treatment_name, mediators_names, outcome_name, backdoor_adjustment
        )
    else:
        estimands_dict["mediation"] = None
    # Finally returning the estimand object
    estimand = IdentifiedEstimand(
        None,
        treatment_variable=treatment_name,
        outcome_variable=outcome_name,
        estimand_type=estimand_type,
        estimands=estimands_dict,
        backdoor_variables=backdoor_variables_dict,
        instrumental_variables=None,
        frontdoor_variables=None,
        mediator_variables=mediators_names,
        mediation_first_stage_confounders=mediation_first_stage_confounders,
        mediation_second_stage_confounders=mediation_second_stage_confounders,
        default_backdoor_id=None,
    )
    return estimand


def identify_nde_effect(
    graph: CausalGraph,
    treatment_name: List[str],
    outcome_name: str,
    backdoor_adjustment: BackdoorAdjustment,
    estimand_type: EstimandType,
    proceed_when_unidentifiable: bool = False,
):
    estimands_dict = {}
    ### 1. FIRST DOING BACKDOOR IDENTIFICATION
    # First, checking if there are any valid backdoor adjustment sets
    backdoor_sets = identify_backdoor(graph, treatment_name, outcome_name, backdoor_adjustment)
    estimands_dict, backdoor_variables_dict = build_backdoor_estimands_dict(
        graph, treatment_name, outcome_name, backdoor_sets, estimands_dict
    )
    # Setting default "backdoor" identification adjustment set
    default_backdoor_id = get_default_backdoor_set_id(graph, treatment_name, outcome_name, backdoor_variables_dict)
    backdoor_variables_dict["backdoor"] = backdoor_variables_dict.get(str(default_backdoor_id), None)

    ### 2. SECOND, CHECKING FOR MEDIATORS
    # Now checking if there are valid mediator variables
    estimands_dict = {}
    mediation_first_stage_confounders = None
    mediation_second_stage_confounders = None
    mediators_names = identify_mediation(graph, treatment_name, outcome_name)
    logger.info("Mediators for treatment and outcome:" + str(mediators_names))
    if len(mediators_names) > 0:
        mediation_estimand_expr = construct_mediation_estimand(
            estimand_type,
            treatment_name,
            outcome_name,
            mediators_names,
        )
        logger.debug("Identified expression = " + str(mediation_estimand_expr))
        estimands_dict["mediation"] = mediation_estimand_expr
        mediation_first_stage_confounders = identify_mediation_first_stage_confounders(
            graph, treatment_name, outcome_name, mediators_names, backdoor_adjustment
        )
        mediation_second_stage_confounders = identify_mediation_second_stage_confounders(
            graph, treatment_name, mediators_names, outcome_name, backdoor_adjustment
        )
    else:
        estimands_dict["mediation"] = None
    # Finally returning the estimand object
    estimand = IdentifiedEstimand(
        None,
        treatment_variable=treatment_name,
        outcome_variable=outcome_name,
        estimand_type=estimand_type,
        estimands=estimands_dict,
        backdoor_variables=backdoor_variables_dict,
        instrumental_variables=None,
        frontdoor_variables=None,
        mediator_variables=mediators_names,
        mediation_first_stage_confounders=mediation_first_stage_confounders,
        mediation_second_stage_confounders=mediation_second_stage_confounders,
        default_backdoor_id=None,
    )
    return estimand


def identify_backdoor(
    graph: CausalGraph,
    treatment_name: List[str],
    outcome_name: str,
    backdoor_adjustment: BackdoorAdjustment,
    include_unobserved: bool = False,
    dseparation_algo: str = "default",
    direct_effect: bool = False,
):
    backdoor_sets = []
    backdoor_paths = None
    bdoor_graph = None
    if dseparation_algo == "naive":
        backdoor_paths = graph.get_backdoor_paths(treatment_name, outcome_name)
    elif dseparation_algo == "default":
        bdoor_graph = graph.do_surgery(
            treatment_name,
            target_node_names=outcome_name,
            remove_outgoing_edges=True,
            remove_only_direct_edges_to_target=direct_effect,
        )
    else:
        raise ValueError(f"d-separation algorithm {dseparation_algo} is not supported")
    backdoor_adjustment = (
        backdoor_adjustment if backdoor_adjustment != BackdoorAdjustment.BACKDOOR_DEFAULT else DEFAULT_BACKDOOR_METHOD
    )

    # First, checking if empty set is a valid backdoor set
    empty_set = set()
    check = graph.check_valid_backdoor_set(
        treatment_name,
        outcome_name,
        empty_set,
        backdoor_paths=backdoor_paths,
        new_graph=bdoor_graph,
        dseparation_algo=dseparation_algo,
    )
    if check["is_dseparated"]:
        backdoor_sets.append({"backdoor_set": empty_set})
        # If the method is `minimal-adjustment`, return the empty set right away.
        if backdoor_adjustment == BackdoorAdjustment.BACKDOOR_MIN:
            return backdoor_sets

    # Second, checking for all other sets of variables. If include_unobserved is false, then only observed variables are eligible.
    eligible_variables = (
        graph.get_all_nodes(include_unobserved=include_unobserved) - set(treatment_name) - set(outcome_name)
    )
    if direct_effect:
        # only remove descendants of Y
        # also allow any causes of Y that are not caused by T (for lower variance)
        eligible_variables -= graph.get_descendants(outcome_name)
    else:
        # remove descendants of T (mediators) and descendants of Y
        eligible_variables -= graph.get_descendants(treatment_name)
    # If var is d-separated from both treatment or outcome, it cannot
    # be a part of the backdoor set
    filt_eligible_variables = set()
    for var in eligible_variables:
        dsep_treat_var = graph.check_dseparation(treatment_name, parse_state(var), set())
        dsep_outcome_var = graph.check_dseparation(outcome_name, parse_state(var), set())
        if not dsep_outcome_var or not dsep_treat_var:
            filt_eligible_variables.add(var)
    if backdoor_adjustment in METHOD_NAMES:
        backdoor_sets, found_valid_adjustment_set = find_valid_adjustment_sets(
            graph,
            treatment_name,
            outcome_name,
            backdoor_paths,
            bdoor_graph,
            dseparation_algo,
            backdoor_sets,
            filt_eligible_variables,
            backdoor_adjustment=backdoor_adjustment,
            max_iterations=MAX_BACKDOOR_ITERATIONS,
        )
        if backdoor_adjustment == BackdoorAdjustment.BACKDOOR_DEFAULT and found_valid_adjustment_set:
            # repeat the above search with BACKDOOR_MIN
            backdoor_sets, _ = find_valid_adjustment_sets(
                graph,
                treatment_name,
                outcome_name,
                backdoor_paths,
                bdoor_graph,
                dseparation_algo,
                backdoor_sets,
                filt_eligible_variables,
                backdoor_adjustment=BackdoorAdjustment.BACKDOOR_MIN,
                max_iterations=MAX_BACKDOOR_ITERATIONS,
            )
    else:
        raise ValueError(
            f"Identifier method {backdoor_adjustment} not supported. Try one of the following: {METHOD_NAMES}"
        )
    return backdoor_sets


def identify_efficient_backdoor(
    graph: CausalGraph,
    backdoor_adjustment: BackdoorAdjustment,
    costs: List,
    conditional_node_names: List[str] = None,
):
    """Method implementing algorithms to compute efficient backdoor sets, as
    described in Rotnitzky and Smucler (2020), Smucler, Sapienza and Rotnitzky (2021)
    and Smucler and Rotnitzky (2022).

    For backdoor_adjustment='efficient-adjustment', computes an optimal backdoor set,
    that is, a backdoor set comprised of observable variables that yields non-parametric
    estimators of the interventional mean with the smallest asymptotic variance
    among those that are based on observable backdoor sets. This optimal backdoor
    set always exists when no variables are latent, and the algorithm is guaranteed to compute
    it in this case. Under a non-parametric graphical model with latent variables,
    such a backdoor set can fail to exist. When certain sufficient conditions under which it is
    known that such a backdoor set exists are not satisfied, an error is raised.

    For backdoor_adjustment='efficient-minimal-adjustment', computes an optimal minimal backdoor set,
    that is, a minimal backdoor set comprised of observable variables that yields non-parametric
    estimators of the interventional mean with the smallest asymptotic variance
    among those that are based on observable minimal backdoor sets.

    For backdoor_adjustment='efficient-mincost-adjustment', computes an optimal minimum cost backdoor set,
    that is, a minimum cost backdoor set comprised of observable variables that yields non-parametric
    estimators of the interventional mean with the smallest asymptotic variance
    among those that are based on observable minimum cost backdoor sets. The cost
    of a backdoor set is defined as the sum of the costs of the variables that comprise it.

    The various optimal backdoor sets computed by this method are not only optimal under
    non-parametric graphical models and non-parametric estimators of interventional mean,
    but also under linear graphical models and OLS estimators, per results in Henckel, Perkovic
    and Maathuis (2020).

    :param costs: a list with non-negative costs associated with variables in the graph. Only used
    for estimatand_type='non-parametric-ate' and backdoor_adjustment='efficient-mincost-adjustment'. If
    not costs are provided by the user, and backdoor_adjustment='efficient-mincost-adjustment', costs
    are assumed to be equal to one for all variables in the graph. The structure of the list should
    be of the form [(node, {"cost": x}) for node in nodes].
    :param conditional_node_names: variables that are used to determine treatment. If none are
    provided, it is assumed that the intervention sets the treatment to a constant.
    :returns:  backdoor_sets, a list of dictionaries, with each dictionary
    having as values a backdoor set.
    """
    if costs is None and backdoor_adjustment == "efficient-mincost-adjustment":
        logger.warning("No costs were passed, so they will be assumed to be constant and equal to 1.")
    efficient_bd = EfficientBackdoor(
        graph=graph,
        conditional_node_names=conditional_node_names,
        costs=costs,
    )
    if backdoor_adjustment == BackdoorAdjustment.BACKDOOR_EFFICIENT:
        backdoor_set = efficient_bd.optimal_adj_set()
        backdoor_sets = [{"backdoor_set": tuple(backdoor_set)}]
    elif backdoor_adjustment == BackdoorAdjustment.BACKDOOR_MIN_EFFICIENT:
        backdoor_set = efficient_bd.optimal_minimal_adj_set()
        backdoor_sets = [{"backdoor_set": tuple(backdoor_set)}]
    elif backdoor_adjustment == BackdoorAdjustment.BACKDOOR_MINCOST_EFFICIENT:
        backdoor_set = efficient_bd.optimal_mincost_adj_set()
        backdoor_sets = [{"backdoor_set": tuple(backdoor_set)}]
    return backdoor_sets


def find_valid_adjustment_sets(
    graph: CausalGraph,
    treatment_name: List,
    outcome_name: List,
    backdoor_paths: List,
    bdoor_graph: CausalGraph,
    dseparation_algo: str,
    backdoor_sets: List,
    filt_eligible_variables: List,
    backdoor_adjustment: BackdoorAdjustment,
    max_iterations: int,
):
    num_iterations = 0
    found_valid_adjustment_set = False
    all_nodes_observed = graph.all_observed(graph.get_all_nodes())
    # If `minimal-adjustment` method is specified, start the search from the set with minimum size. Otherwise, start from the largest.
    set_sizes = (
        range(1, len(filt_eligible_variables) + 1, 1)
        if backdoor_adjustment == BackdoorAdjustment.BACKDOOR_MIN
        else range(len(filt_eligible_variables), 0, -1)
    )
    for size_candidate_set in set_sizes:
        for candidate_set in itertools.combinations(filt_eligible_variables, size_candidate_set):
            check = graph.check_valid_backdoor_set(
                treatment_name,
                outcome_name,
                candidate_set,
                backdoor_paths=backdoor_paths,
                new_graph=bdoor_graph,
                dseparation_algo=dseparation_algo,
            )
            logger.debug(
                "Candidate backdoor set: {0}, is_dseparated: {1}".format(candidate_set, check["is_dseparated"])
            )
            if check["is_dseparated"]:
                backdoor_sets.append({"backdoor_set": candidate_set})
                found_valid_adjustment_set = True
            num_iterations += 1
            if backdoor_adjustment == BackdoorAdjustment.BACKDOOR_EXHAUSTIVE and num_iterations > max_iterations:
                logger.warning(f"Max number of iterations {max_iterations} reached.")
                break
        # If the backdoor method is `maximal-adjustment` or `minimal-adjustment`, return the first found adjustment set.
        if (
            backdoor_adjustment
            in {
                BackdoorAdjustment.BACKDOOR_DEFAULT,
                BackdoorAdjustment.BACKDOOR_MAX,
                BackdoorAdjustment.BACKDOOR_MIN,
            }
            and found_valid_adjustment_set
        ):
            break
        # If all variables are observed, and the biggest eligible set
        # does not satisfy backdoor, then none of its subsets will.
        if (
            backdoor_adjustment in {BackdoorAdjustment.BACKDOOR_DEFAULT, BackdoorAdjustment.BACKDOOR_MAX}
            and all_nodes_observed
        ):
            break
        if num_iterations > max_iterations:
            logger.warning(f"Max number of iterations {max_iterations} reached. Could not find a valid backdoor set.")
            break
    return backdoor_sets, found_valid_adjustment_set


def get_default_backdoor_set_id(
    graph: CausalGraph, treatment_name: List[str], outcome_name: List[str], backdoor_sets_dict: Dict
):
    # Adding a None estimand if no backdoor set found
    if len(backdoor_sets_dict) == 0:
        return None

    # Default set contains minimum possible number of instrumental variables, to prevent lowering variance in the treatment variable.
    instrument_names = set(graph.get_instruments(treatment_name, outcome_name))
    iv_count_dict = {
        key: len(set(bdoor_set).intersection(instrument_names)) for key, bdoor_set in backdoor_sets_dict.items()
    }
    min_iv_count = min(iv_count_dict.values())
    min_iv_keys = {key for key, iv_count in iv_count_dict.items() if iv_count == min_iv_count}
    min_iv_backdoor_sets_dict = {key: backdoor_sets_dict[key] for key in min_iv_keys}

    # Default set is the one with the least number of adjustment variables (optimizing for efficiency)
    min_set_length = 1000000
    default_key = None
    for key, bdoor_set in min_iv_backdoor_sets_dict.items():
        if len(bdoor_set) < min_set_length:
            min_set_length = len(bdoor_set)
            default_key = key
    return default_key


def build_backdoor_estimands_dict(
    graph: CausalGraph,
    treatment_name: List[str],
    outcome_name: List[str],
    backdoor_sets: List[str],
    estimands_dict: Dict,
):
    """Build the final dict for backdoor sets by filtering unobserved variables if needed."""
    backdoor_variables_dict = {}
    is_identified = [graph.all_observed(bset["backdoor_set"]) for bset in backdoor_sets]

    if any(is_identified):
        logger.info("Causal effect can be identified.")
        backdoor_sets_arr = [
            list(bset["backdoor_set"]) for bset in backdoor_sets if graph.all_observed(bset["backdoor_set"])
        ]
    else:  # there is unobserved confounding
        logger.warning("Backdoor identification failed.")
        backdoor_sets_arr = []

    for i in range(len(backdoor_sets_arr)):
        backdoor_estimand_expr = construct_backdoor_estimand(treatment_name, outcome_name, backdoor_sets_arr[i])
        logger.debug("Identified expression = " + str(backdoor_estimand_expr))
        estimands_dict["backdoor" + str(i + 1)] = backdoor_estimand_expr
        backdoor_variables_dict["backdoor" + str(i + 1)] = backdoor_sets_arr[i]
    return estimands_dict, backdoor_variables_dict


def identify_frontdoor(
    graph: CausalGraph, treatment_name: List[str], outcome_name: List[str], dseparation_algo: str = "default"
):
    """Find a valid frontdoor variable if it exists.

    Currently only supports a single variable frontdoor set.
    """
    frontdoor_var = None
    frontdoor_paths = None
    fdoor_graph = None
    if dseparation_algo == "default":
        cond1_graph = graph.do_surgery(treatment_name, remove_incoming_edges=True)
        bdoor_graph1 = graph.do_surgery(treatment_name, remove_outgoing_edges=True)
    elif dseparation_algo == "naive":
        frontdoor_paths = graph.get_all_directed_paths(treatment_name, outcome_name)
    else:
        raise ValueError(f"d-separation algorithm {dseparation_algo} is not supported")

    eligible_variables = (
        graph.get_descendants(treatment_name) - set(outcome_name) - set(graph.get_descendants(outcome_name))
    )
    # For simplicity, assuming a one-variable frontdoor set
    for candidate_var in eligible_variables:
        # Cond 1: All directed paths intercepted by candidate_var
        cond1 = graph.check_valid_frontdoor_set(
            treatment_name,
            outcome_name,
            parse_state(candidate_var),
            frontdoor_paths=frontdoor_paths,
            new_graph=cond1_graph,
            dseparation_algo=dseparation_algo,
        )
        logger.debug("Candidate frontdoor set: {0}, is_dseparated: {1}".format(candidate_var, cond1))
        if not cond1:
            continue
        # Cond 2: No confounding between treatment and candidate var
        cond2 = graph.check_valid_backdoor_set(
            treatment_name,
            parse_state(candidate_var),
            set(),
            backdoor_paths=None,
            new_graph=bdoor_graph1,
            dseparation_algo=dseparation_algo,
        )
        if not cond2:
            continue
        # Cond 3: treatment blocks all confounding between candidate_var and outcome
        bdoor_graph2 = graph.do_surgery(candidate_var, remove_outgoing_edges=True)
        cond3 = graph.check_valid_backdoor_set(
            parse_state(candidate_var),
            outcome_name,
            treatment_name,
            backdoor_paths=None,
            new_graph=bdoor_graph2,
            dseparation_algo=dseparation_algo,
        )
        is_valid_frontdoor = cond1 and cond2 and cond3
        if is_valid_frontdoor:
            frontdoor_var = candidate_var
            break
    return parse_state(frontdoor_var)


def identify_mediation(graph: CausalGraph, treatment_name: List[str], outcome_name: List[str]):
    """Find a valid mediator if it exists.

    Currently only supports a single variable mediator set.
    """
    mediation_var = None
    mediation_paths = graph.get_all_directed_paths(treatment_name, outcome_name)
    eligible_variables = graph.get_descendants(treatment_name) - set(outcome_name)
    # For simplicity, assuming a one-variable mediation set
    for candidate_var in eligible_variables:
        is_valid_mediation = graph.check_valid_mediation_set(
            treatment_name,
            outcome_name,
            parse_state(candidate_var),
            mediation_paths=mediation_paths,
        )
        logger.debug("Candidate mediation set: {0}, on_mediating_path: {1}".format(candidate_var, is_valid_mediation))
        if is_valid_mediation:
            mediation_var = candidate_var
            break
    return parse_state(mediation_var)


def identify_mediation_first_stage_confounders(
    graph: CausalGraph,
    treatment_name: List[str],
    outcome_name: List[str],
    mediators_names: List[str],
    backdoor_adjustment: BackdoorAdjustment,
):
    # Create estimands dict as per the API for backdoor, but do not return it
    estimands_dict = {}
    backdoor_sets = identify_backdoor(graph, treatment_name, mediators_names, backdoor_adjustment)
    estimands_dict, backdoor_variables_dict = build_backdoor_estimands_dict(
        graph,
        treatment_name,
        mediators_names,
        backdoor_sets,
        estimands_dict,
    )
    # Setting default "backdoor" identification adjustment set
    default_backdoor_id = get_default_backdoor_set_id(graph, treatment_name, outcome_name, backdoor_variables_dict)
    estimands_dict["backdoor"] = estimands_dict.get(str(default_backdoor_id), None)
    backdoor_variables_dict["backdoor"] = backdoor_variables_dict.get(str(default_backdoor_id), None)
    return backdoor_variables_dict


def identify_mediation_second_stage_confounders(
    graph: CausalGraph,
    treatment_name: List[str],
    mediators_names: List[str],
    outcome_name: List[str],
    backdoor_adjustment: BackdoorAdjustment,
):
    # Create estimands dict as per the API for backdoor, but do not return it
    estimands_dict = {}
    backdoor_sets = identify_backdoor(graph, mediators_names, outcome_name, backdoor_adjustment)
    estimands_dict, backdoor_variables_dict = build_backdoor_estimands_dict(
        graph,
        mediators_names,
        outcome_name,
        backdoor_sets,
        estimands_dict,
    )
    # Setting default "backdoor" identification adjustment set
    default_backdoor_id = get_default_backdoor_set_id(graph, treatment_name, outcome_name, backdoor_variables_dict)
    estimands_dict["backdoor"] = estimands_dict.get(str(default_backdoor_id), None)
    backdoor_variables_dict["backdoor"] = backdoor_variables_dict.get(str(default_backdoor_id), None)
    return backdoor_variables_dict


def construct_backdoor_estimand(treatment_name: List[str], outcome_name: List[str], common_causes: List[str]):
    # TODO: outputs string for now, but ideally should do symbolic
    # expressions Mon 19 Feb 2018 04:54:17 PM DST
    # TODO Better support for multivariate treatments

    expr = None
    outcome_name = outcome_name[0]
    num_expr_str = outcome_name
    if len(common_causes) > 0:
        num_expr_str += "|" + ",".join(common_causes)
    expr = "d(" + num_expr_str + ")/d" + ",".join(treatment_name)
    sym_mu = sp.Symbol("mu")
    sym_sigma = sp.Symbol("sigma", positive=True)
    sym_outcome = spstats.Normal(num_expr_str, sym_mu, sym_sigma)
    sym_treatment_symbols = [sp.Symbol(t) for t in treatment_name]
    sym_treatment = sp.Array(sym_treatment_symbols)
    sym_conditional_outcome = spstats.Expectation(sym_outcome)
    sym_effect = sp.Derivative(sym_conditional_outcome, sym_treatment)

    sym_assumptions = {
        "Unconfoundedness": (
            "If U\N{RIGHTWARDS ARROW}{{{0}}} and U\N{RIGHTWARDS ARROW}{1}" " then P({1}|{0},{2},U) = P({1}|{0},{2})"
        ).format(",".join(treatment_name), outcome_name, ",".join(common_causes))
    }

    estimand = {"estimand": sym_effect, "assumptions": sym_assumptions}
    return estimand


def construct_iv_estimand(treatment_name: List[str], outcome_name: List[str], instrument_names: List[str]):
    # TODO: support multivariate treatments better.
    expr = None
    outcome_name = outcome_name[0]
    sym_outcome = spstats.Normal(outcome_name, 0, 1)
    sym_treatment_symbols = [spstats.Normal(t, 0, 1) for t in treatment_name]
    sym_treatment = sp.Array(sym_treatment_symbols)
    sym_instrument_symbols = [sp.Symbol(inst) for inst in instrument_names]
    sym_instrument = sp.Array(sym_instrument_symbols)  # ",".join(instrument_names))
    sym_outcome_derivative = sp.Derivative(sym_outcome, sym_instrument)
    sym_treatment_derivative = sp.Derivative(sym_treatment, sym_instrument)
    sym_effect = spstats.Expectation(sym_outcome_derivative / sym_treatment_derivative)
    sym_assumptions = {
        "As-if-random": (
            "If U\N{RIGHTWARDS ARROW}\N{RIGHTWARDS ARROW}{0} then "
            "\N{NOT SIGN}(U \N{RIGHTWARDS ARROW}\N{RIGHTWARDS ARROW}{{{1}}})"
        ).format(outcome_name, ",".join(instrument_names)),
        "Exclusion": (
            "If we remove {{{0}}}\N{RIGHTWARDS ARROW}{{{1}}}, then " "\N{NOT SIGN}({{{0}}}\N{RIGHTWARDS ARROW}{2})"
        ).format(",".join(instrument_names), ",".join(treatment_name), outcome_name),
    }

    estimand = {"estimand": sym_effect, "assumptions": sym_assumptions}
    return estimand


def construct_frontdoor_estimand(
    treatment_name: List[str], outcome_name: List[str], frontdoor_variables_names: List[str]
):
    # TODO: support multivariate treatments better.
    expr = None
    outcome_name = outcome_name[0]
    sym_outcome = spstats.Normal(outcome_name, 0, 1)
    sym_treatment_symbols = [spstats.Normal(t, 0, 1) for t in treatment_name]
    sym_treatment = sp.Array(sym_treatment_symbols)
    sym_frontdoor_symbols = [sp.Symbol(inst) for inst in frontdoor_variables_names]
    sym_frontdoor = sp.Array(sym_frontdoor_symbols)  # ",".join(instrument_names))
    sym_outcome_derivative = sp.Derivative(sym_outcome, sym_frontdoor)
    sym_treatment_derivative = sp.Derivative(sym_frontdoor, sym_treatment)
    sym_effect = spstats.Expectation(sym_treatment_derivative * sym_outcome_derivative)
    sym_assumptions = {
        "Full-mediation": ("{2} intercepts (blocks) all directed paths from {0} to {1}.").format(
            ",".join(treatment_name),
            ",".join(outcome_name),
            ",".join(frontdoor_variables_names),
        ),
        "First-stage-unconfoundedness": (
            "If U\N{RIGHTWARDS ARROW}{{{0}}} and U\N{RIGHTWARDS ARROW}{{{1}}}" " then P({1}|{0},U) = P({1}|{0})"
        ).format(",".join(treatment_name), ",".join(frontdoor_variables_names)),
        "Second-stage-unconfoundedness": (
            "If U\N{RIGHTWARDS ARROW}{{{2}}} and U\N{RIGHTWARDS ARROW}{1}" " then P({1}|{2}, {0}, U) = P({1}|{2}, {0})"
        ).format(
            ",".join(treatment_name),
            outcome_name,
            ",".join(frontdoor_variables_names),
        ),
    }

    estimand = {"estimand": sym_effect, "assumptions": sym_assumptions}
    return estimand


def construct_mediation_estimand(
    estimand_type: EstimandType, treatment_name: List[str], outcome_name: List[str], mediators_names: List[str]
):
    # TODO: support multivariate treatments better.
    expr = None
    if estimand_type in (
        EstimandType.NONPARAMETRIC_NDE,
        EstimandType.NONPARAMETRIC_NIE,
    ):
        outcome_name = outcome_name[0]
        sym_outcome = spstats.Normal(outcome_name, 0, 1)
        sym_treatment_symbols = [spstats.Normal(t, 0, 1) for t in treatment_name]
        sym_treatment = sp.Array(sym_treatment_symbols)
        sym_mediators_symbols = [sp.Symbol(inst) for inst in mediators_names]
        sym_mediators = sp.Array(sym_mediators_symbols)
        sym_outcome_derivative = sp.Derivative(sym_outcome, sym_mediators)
        sym_treatment_derivative = sp.Derivative(sym_mediators, sym_treatment)
        # For direct effect
        num_expr_str = outcome_name
        if len(mediators_names) > 0:
            num_expr_str += "|" + ",".join(mediators_names)
        sym_mu = sp.Symbol("mu")
        sym_sigma = sp.Symbol("sigma", positive=True)
        sym_conditional_outcome = spstats.Normal(num_expr_str, sym_mu, sym_sigma)
        sym_directeffect_derivative = sp.Derivative(sym_conditional_outcome, sym_treatment)
        if estimand_type == EstimandType.NONPARAMETRIC_NIE:
            sym_effect = spstats.Expectation(sym_treatment_derivative * sym_outcome_derivative)
        elif estimand_type == EstimandType.NONPARAMETRIC_NDE:
            sym_effect = spstats.Expectation(sym_directeffect_derivative)
        sym_assumptions = {
            "Mediation": (
                "{2} intercepts (blocks) all directed paths from {0} to {1} except the path {{{0}}}\N{RIGHTWARDS ARROW}{{{1}}}."
            ).format(
                ",".join(treatment_name),
                ",".join(outcome_name),
                ",".join(mediators_names),
            ),
            "First-stage-unconfoundedness": (
                "If U\N{RIGHTWARDS ARROW}{{{0}}} and U\N{RIGHTWARDS ARROW}{{{1}}}" " then P({1}|{0},U) = P({1}|{0})"
            ).format(",".join(treatment_name), ",".join(mediators_names)),
            "Second-stage-unconfoundedness": (
                "If U\N{RIGHTWARDS ARROW}{{{2}}} and U\N{RIGHTWARDS ARROW}{1}"
                " then P({1}|{2}, {0}, U) = P({1}|{2}, {0})"
            ).format(",".join(treatment_name), outcome_name, ",".join(mediators_names)),
        }
    else:
        raise ValueError(
            "Estimand type not supported. Supported estimand types are {0} or {1}'.".format(
                EstimandType.NONPARAMETRIC_NDE,
                EstimandType.NONPARAMETRIC_NIE,
            )
        )

    estimand = {"estimand": sym_effect, "assumptions": sym_assumptions}
    return estimand



import numpy as np
import pytest
from pytest import mark

from .base import TestRefuter


@mark.usefixtures("fixed_seed")
class TestDataSubsetRefuter(object):
    """
    The first two tests are for the default behavior, in which we just bootstrap the data
    and obtain the estimate.

    """

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "num_samples"], [(0.05, "iv.instrumental_variable", 1000)]
    )
    def test_refutation_bootstrap_refuter_continuous(self, error_tolerance, estimator_method, num_samples):
        refuter_tester = TestRefuter(error_tolerance, estimator_method, "bootstrap_refuter")
        refuter_tester.continuous_treatment_testsuite(num_samples=num_samples)  # Run both

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "num_samples"], [(0.05, "backdoor.propensity_score_matching", 1000)]
    )
    def test_refutation_bootstrap_refuter_binary(self, error_tolerance, estimator_method, num_samples):
        refuter_tester = TestRefuter(error_tolerance, estimator_method, "bootstrap_refuter")
        refuter_tester.binary_treatment_testsuite(tests_to_run="atleast-one-common-cause", num_samples=num_samples)

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "num_common_causes", "required_variables", "num_samples"],
        [(0.05, "iv.instrumental_variable", 5, 3, 1000)],
    )
    def test_refutation_bootstrap_refuter_continuous_integer_argument(
        self, error_tolerance, estimator_method, num_common_causes, required_variables, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerance,
            estimator_method,
            "bootstrap_refuter",
            required_variables=required_variables,
        )
        refuter_tester.continuous_treatment_testsuite(
            num_samples=num_samples, num_common_causes=num_common_causes, tests_to_run="atleast-one-common-cause"
        )  # Run atleast one common cause

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "num_common_causes", "required_variables", "num_samples"],
        [(0.05, "iv.instrumental_variable", 5, ["W0", "W1"], 1000)],
    )
    def test_refutation_bootstrap_refuter_continuous_list_argument(
        self, error_tolerance, estimator_method, num_common_causes, required_variables, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerance, estimator_method, "bootstrap_refuter", required_variables=required_variables
        )
        refuter_tester.continuous_treatment_testsuite(
            num_samples=num_samples, num_common_causes=num_common_causes, tests_to_run="atleast-one-common-cause"
        )  # Run atleast one common cause

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "num_common_causes", "required_variables", "num_samples"],
        [(0.1, "backdoor.propensity_score_matching", 5, 3, 5000)],
    )
    def test_refutation_bootstrap_refuter_binary_integer_argument(
        self, error_tolerance, estimator_method, num_common_causes, required_variables, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerance, estimator_method, "bootstrap_refuter", required_variables=required_variables
        )
        refuter_tester.binary_treatment_testsuite(
            num_samples=num_samples, num_common_causes=num_common_causes, tests_to_run="atleast-one-common-cause"
        )

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "num_common_causes", "required_variables", "num_samples"],
        [(0.1, "backdoor.propensity_score_matching", 5, ["W0", "W1"], 5000)],
    )
    def test_refutation_bootstrap_refuter_binary_list_argument(
        self, error_tolerance, estimator_method, num_common_causes, required_variables, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerance, estimator_method, "bootstrap_refuter", required_variables=required_variables
        )
        refuter_tester.binary_treatment_testsuite(
            num_samples=num_samples, num_common_causes=num_common_causes, tests_to_run="atleast-one-common-cause"
        )

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "num_common_causes", "required_variables", "num_samples"],
        [(0.1, "iv.instrumental_variable", 5, ["-W0", "-W1"], 5000)],
    )
    def test_refutation_bootstrap_refuter_continuous_list_negative_argument(
        self, error_tolerance, estimator_method, num_common_causes, required_variables, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerance, estimator_method, "bootstrap_refuter", required_variables=required_variables
        )
        refuter_tester.continuous_treatment_testsuite(
            num_samples=num_samples, num_common_causes=num_common_causes, tests_to_run="atleast-one-common-cause"
        )  # Run atleast one common cause

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "num_common_causes", "required_variables", "num_samples"],
        [(0.1, "backdoor.propensity_score_matching", 5, ["-W0", "-W1"], 5000)],
    )
    def test_refutation_bootstrap_refuter_binary_list_negative_argument(
        self, error_tolerance, estimator_method, num_common_causes, required_variables, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerance, estimator_method, "bootstrap_refuter", required_variables=required_variables
        )
        refuter_tester.binary_treatment_testsuite(
            num_samples=num_samples, num_common_causes=num_common_causes, tests_to_run="atleast-one-common-cause"
        )



import logging
import sys

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import statsmodels.api as sm
from scipy.stats import t

from dowhy.utils.api import parse_state


class LinearSensitivityAnalyzer:
    """
    Class to perform sensitivity analysis
    See: https://carloscinelli.com/files/Cinelli%20and%20Hazlett%20(2020)%20-%20Making%20Sense%20of%20Sensitivity.pdf

    :param estimator: linear estimator of the causal model
    :param data: Pandas dataframe
    :param treatment_name: name of treatment
     :param percent_change_estimate: It is the percentage of reduction of treatment estimate that could alter the results (default = 1)
                                        if percent_change_estimate = 1, the robustness value describes the strength of association of confounders with treatment and outcome in order to reduce the estimate by 100% i.e bring it down to 0.
    :param null_hypothesis_effect: assumed effect under the null hypothesis
    :param confounder_increases_estimate: True implies that confounder increases the absolute value of estimate and vice versa. (Default = True)
    :param benchmark_common_causes: names of variables for bounding strength of confounders
    :param significance_level: confidence interval for statistical inference(default = 0.05)
    :param frac_strength_treatment: strength of association between unobserved confounder and treatment compared to benchmark covariate
    :param frac_strength_outcome: strength of association between unobserved confounder and outcome compared to benchmark covariate
    :param common_causes_order: The order of column names in OLS regression data
    """

    def __init__(
        self,
        estimator=None,
        data=None,
        treatment_name=None,
        percent_change_estimate=1.0,
        significance_level=0.05,
        confounder_increases_estimate=True,
        benchmark_common_causes=None,
        null_hypothesis_effect=0,
        frac_strength_treatment=None,
        frac_strength_outcome=None,
        common_causes_order=None,
    ):
        self.data = data
        self.treatment_name = []
        # original_treatment_name: : stores original variable names for labelling
        self.original_treatment_name = treatment_name
        for t in range(len(treatment_name)):
            self.treatment_name.append("x" + str(t + 1))

        self.percent_change_estimate = percent_change_estimate
        self.significance_level = significance_level
        self.confounder_increases_estimate = confounder_increases_estimate
        self.estimator = estimator
        self.estimator_model = estimator.model
        self.null_hypothesis_effect = null_hypothesis_effect

        # common_causes_map : maps the original variable names to variable names in OLS regression
        self.common_causes_map = {}
        for i in range(len(common_causes_order)):
            self.common_causes_map[common_causes_order[i]] = "x" + str(len(self.treatment_name) + i + 1)

        # benchmark_common_causes: stores variable names in terms of regression model variables
        benchmark_common_causes = parse_state(benchmark_common_causes)
        self.benchmark_common_causes = []
        # original_benchmark_covariates: stores original variable names for labelling
        self.original_benchmark_covariates = benchmark_common_causes
        for i in range(len(benchmark_common_causes)):
            self.benchmark_common_causes.append(self.common_causes_map[benchmark_common_causes[i]])

        if type(frac_strength_treatment) in [int, list, float]:
            self.frac_strength_treatment = np.array(frac_strength_treatment)
        if type(frac_strength_outcome) in [int, list, float]:
            self.frac_strength_outcome = np.array(frac_strength_outcome)

        # estimate: estimate of regression
        self.estimate = None
        # degree_of_freedom: degree of freedom of error in regression
        self.degree_of_freedom = None
        # standard_error: standard error in regression
        self.standard_error = None
        # t_stats: Treatment coefficient t-value - measures how many standard errors the estimate is away from zero.
        self.t_stats = None
        # partial_f2: value to determine if a regression model and a nested version of it have a statistically significant difference between them
        self.partial_f2 = None
        # r2tu_w: partial R^2  of unobserved confounder "u" with treatment "t", after conditioning on observed covariates "w"
        self.r2tu_w = None
        # r2yu_tw: partial R^2  of unobserved confounder "u" with outcome "y", after conditioning on observed covariates "w" and treatment "t"
        self.r2yu_tw = None
        # r2twj_w: partial R^2 of observed covariate wj with treatment "t", after conditioning on observed covariates "w" excluding wj
        self.r2twj_w = None
        # r2ywj_tw:  partial R^2 of observed covariate wj with outcome "y", after conditioning on observed covariates "w" (excluding wj) and treatment "t"
        self.r2ywj_tw = None
        # benchmarking_results: dataframe containing information about bounds and bias adjusted terms
        self.benchmarking_results = None
        # stats: dictionary containing information like robustness value, partial R^2, estimate, standard error , degree of freedom, partial f^2, t-statistic
        self.stats = None
        self.logger = logging.getLogger(__name__)

    def treatment_regression(self):
        """
        Function to perform regression with treatment as outcome

        :returns: new OLS regression model
        """

        features = self.estimator._observed_common_causes.copy()
        treatment_df = self.estimator._treatment.copy()
        features = sm.tools.add_constant(features)
        features.rename(columns=self.common_causes_map, inplace=True)
        model = sm.OLS(treatment_df, features)
        estimator_model = model.fit()

        return estimator_model

    def partial_r2_func(self, estimator_model=None, treatment=None):
        """
        Computes the partial R^2 of regression model

        :param estimator_model: Linear regression model
        :param treatment: treatment name

        :returns: partial R^2 value
        """

        estimate = estimator_model.params[treatment]
        degree_of_freedom = int(estimator_model.df_resid)

        if np.isscalar(estimate):  # for single covariate
            t_stats = estimator_model.tvalues[treatment]
            return t_stats**2 / (t_stats**2 + degree_of_freedom)

        else:  # compute for a group of covariates
            covariance_matrix = estimator_model.cov_params().loc[treatment, :][treatment]
            n = len(estimate)  # number of parameters in model
            f_stat = (
                np.matmul(np.matmul(estimate.values.T, np.linalg.inv(covariance_matrix.values)), estimate.values) / n
            )
            return f_stat * n / (f_stat * n + degree_of_freedom)

    def robustness_value_func(self, alpha=1.0):
        """
        Function to calculate the robustness value.
        It is the minimum strength of association that confounders must have with treatment and outcome to change conclusions.
        Robustness value describes how strong the association must be in order to reduce the estimated effect by (100 * percent_change_estimate)%.
        Robustness value close to 1 means the treatment effect can handle strong confounders explaining  almost all residual variation of the treatment and the outcome.
        Robustness value close to 0 means that even very weak confounders can also change the results.

        :param alpha: confidence interval (default = 1)

        :returns: robustness value
        """

        partial_cohen_f = abs(
            self.t_stats / np.sqrt(self.degree_of_freedom)
        )  # partial f of treatment t with outcome y. f = t_val/sqrt(dof)
        f_q = self.percent_change_estimate * partial_cohen_f
        t_alpha_df_1 = t.ppf(
            alpha / 2, self.degree_of_freedom - 1
        )  # t-value threshold with alpha significance level and dof-1 degrees of freedom
        f_critical = abs(t_alpha_df_1) / np.sqrt(self.degree_of_freedom - 1)
        f_adjusted = f_q - f_critical

        if f_adjusted < 0:
            r_value = 0
        else:
            r_value = 0.5 * (np.sqrt(f_adjusted**4 + (4 * f_adjusted**2)) - f_adjusted**2)

        if f_adjusted > 0 and f_q > 1 / f_critical:
            r_value = (f_q**2 - f_critical**2) / (1 + f_q**2)

        return r_value

    def compute_bias_adjusted(self, r2tu_w, r2yu_tw):
        """
        Computes the bias adjusted estimate, standard error, t-value,  partial R2, confidence intervals

        :param r2tu_w: partial r^2 from regressing unobserved confounder u on treatment t after conditioning on observed covariates w
        :param r2yu_tw: partial r^2 from regressing unobserved confounder u on outcome y after conditioning on observed covariates w and treatment t

        :returns: Python dictionary with information about partial R^2 of confounders with treatment and outcome and bias adjusted variables
        """

        bias_factor = np.sqrt((r2yu_tw * r2tu_w) / (1 - r2tu_w))
        bias = bias_factor * (self.standard_error * np.sqrt(self.degree_of_freedom))

        if self.confounder_increases_estimate:
            bias_adjusted_estimate = np.sign(self.estimate) * (abs(self.estimate) - bias)
        else:
            bias_adjusted_estimate = np.sign(self.estimate) * (abs(self.estimate) + bias)

        bias_adjusted_se = (
            np.sqrt((1 - r2yu_tw) / (1 - r2tu_w))
            * self.standard_error
            * np.sqrt(self.degree_of_freedom / (self.degree_of_freedom - 1))
        )

        bias_adjusted_t = (bias_adjusted_estimate - self.null_hypothesis_effect) / bias_adjusted_se

        bias_adjusted_partial_r2 = bias_adjusted_t**2 / (
            bias_adjusted_t**2 + (self.degree_of_freedom - 1)
        )  # partial r2 formula used with new t value and dof - 1

        num_se = t.ppf(
            self.significance_level / 2, self.degree_of_freedom
        )  # Number of standard errors within Confidence Interval

        bias_adjusted_upper_CI = bias_adjusted_estimate - num_se * bias_adjusted_se
        bias_adjusted_lower_CI = bias_adjusted_estimate + num_se * bias_adjusted_se

        benchmarking_results = {
            "r2tu_w": r2tu_w,
            "r2yu_tw": r2yu_tw,
            "bias_adjusted_estimate": bias_adjusted_estimate,
            "bias_adjusted_se": bias_adjusted_se,
            "bias_adjusted_t": bias_adjusted_t,
            "bias_adjusted_lower_CI": bias_adjusted_lower_CI,
            "bias_adjusted_upper_CI": bias_adjusted_upper_CI,
        }

        return benchmarking_results

    def check_sensitivity(self, plot=True):
        """
        Function to perform sensitivity analysis.
        :param plot: plot = True generates a plot of point estimate and the variations with respect to unobserved confounding.
                     plot = False overrides the setting

        :returns: instance of LinearSensitivityAnalyzer class
        """

        self.standard_error = np.array(self.estimator_model.bse[1 : (len(self.treatment_name) + 1)])[0]
        self.degree_of_freedom = int(self.estimator_model.df_resid)
        self.estimate = np.array(self.estimator_model.params[1 : (len(self.treatment_name) + 1)])[0]
        self.t_stats = np.array(self.estimator_model.tvalues[self.treatment_name])[0]

        # partial R^2 (r2yt_w) is the proportion of variation in outcome uniquely explained by treatment
        partial_r2 = self.partial_r2_func(self.estimator_model, self.treatment_name)
        RVq = self.robustness_value_func()
        RV_qalpha = self.robustness_value_func(alpha=self.significance_level)

        if self.confounder_increases_estimate:
            self.null_hypothesis_effect = self.estimate * (1 - self.percent_change_estimate)
        else:
            self.null_hypothesis_effect = self.estimate * (1 + self.percent_change_estimate)

        self.t_stats = (self.estimate - self.null_hypothesis_effect) / self.standard_error
        self.partial_f2 = self.t_stats**2 / self.degree_of_freedom

        # build a new regression model by considering treatment variables as outcome
        treatment_linear_model = self.treatment_regression()

        # r2twj_w is partial R^2 of covariate wj with treatment "t", after conditioning on covariates w(excluding wj)
        # r2ywj_tw is partial R^2 of covariate wj with outcome "y", after conditioning on covariates w(excluding wj) and treatment "t"
        self.r2twj_w = []
        self.r2ywj_tw = []

        for covariate in self.benchmark_common_causes:
            self.r2ywj_tw.append(self.partial_r2_func(self.estimator_model, covariate))
            self.r2twj_w.append(self.partial_r2_func(treatment_linear_model, covariate))

        for i in range(len(self.benchmark_common_causes)):
            r2twj_w = self.r2twj_w[i]
            r2ywj_tw = self.r2ywj_tw[i]

            # r2tu_w is the partial r^2 from regressing u on t after conditioning on w
            self.r2tu_w = self.frac_strength_treatment * (r2twj_w / (1 - r2twj_w))
            if any(val >= 1 for val in self.r2tu_w):
                raise ValueError("r2tu_w can not be >= 1. Try a lower frac_strength_treatment value")

            r2uwj_wt = (
                self.frac_strength_treatment
                * (r2twj_w**2)
                / ((1 - self.frac_strength_treatment * r2twj_w) * (1 - r2twj_w))
            )
            if any(val >= 1 for val in r2uwj_wt):
                raise ValueError("r2uwj_wt can not be >= 1. Try a lower frac_strength_treatment value")

            self.r2yu_tw = ((np.sqrt(self.frac_strength_outcome) + np.sqrt(r2uwj_wt)) / np.sqrt(1 - r2uwj_wt)) ** 2 * (
                r2ywj_tw / (1 - r2ywj_tw)
            )
            if any(val > 1 for val in self.r2yu_tw):
                for i in range(len(self.r2yu_tw)):
                    if self.r2yu_tw[i] > 1:
                        self.r2yu_tw[i] = 1
                self.logger.warning(
                    "Warning: r2yu_tw can not be > 1. Try a lower frac_strength_treatment. Setting r2yu_tw to 1"
                )

            # Compute bias adjusted terms

        self.benchmarking_results = self.compute_bias_adjusted(self.r2tu_w, self.r2yu_tw)

        if plot == True:
            self.plot()

        self.stats = {
            "estimate": self.estimate,
            "standard_error": self.standard_error,
            "degree of freedom": self.degree_of_freedom,
            "t_statistic": self.t_stats,
            "r2yt_w": partial_r2,
            "partial_f2": self.partial_f2,
            "robustness_value": RVq,
            "robustness_value_alpha": RV_qalpha,
        }

        self.benchmarking_results = pd.DataFrame.from_dict(self.benchmarking_results)
        return self

    def plot_estimate(self, r2tu_w, r2yu_tw):
        """
        Computes the contours, threshold line and bounds for plotting estimates.
        Contour lines (z - axis) correspond to the adjusted estimate values for different values of r2tu_w (x) and r2yu_tw (y).
        :param r2tu_w: hypothetical partial R^2 of confounder with treatment(x - axis)
        :param r2yu_tw: hypothetical partial R^2 of confounder with outcome(y - axis)

        :returns:
        contour_values : values of contour lines for the plot
        critical_estimate : threshold point
        estimate_bounds : estimate values for unobserved confounders (bias adjusted estimates)
        """

        critical_estimate = self.null_hypothesis_effect
        contour_values = np.zeros((len(r2yu_tw), len(r2tu_w)))
        for i in range(len(r2yu_tw)):
            y = r2yu_tw[i]
            for j in range(len(r2tu_w)):
                x = r2tu_w[j]
                benchmarking_results = self.compute_bias_adjusted(r2tu_w=x, r2yu_tw=y)
                estimate = benchmarking_results["bias_adjusted_estimate"]
                contour_values[i][j] = estimate

        estimate_bounds = self.benchmarking_results["bias_adjusted_estimate"]
        return contour_values, critical_estimate, estimate_bounds

    def plot_t(self, r2tu_w, r2yu_tw):
        """
        Computes the contours, threshold line and bounds for plotting t.
        Contour lines (z - axis) correspond to the adjusted t values for different values of r2tu_w (x) and r2yu_tw (y).
        :param r2tu_w: hypothetical partial R^2 of confounder with treatment(x - axis)
        :param r2yu_tw: hypothetical partial R^2 of confounder with outcome(y - axis)

        :returns:
        contour_values : values of contour lines for the plot
        critical_t : threshold point
        t_bounds : t-value for unobserved confounders (bias adjusted t values)
        """

        t_alpha_df_1 = t.ppf(
            self.significance_level / 2, self.degree_of_freedom - 1
        )  # t-value threshold with alpha significance level and dof-1 degrees of freedom
        critical_t = abs(t_alpha_df_1) * np.sign(self.t_stats)

        contour_values = []
        for x in r2tu_w:
            contour = []
            for y in r2yu_tw:
                benchmarking_results = self.compute_bias_adjusted(r2tu_w=x, r2yu_tw=y)
                t_value = benchmarking_results["bias_adjusted_t"]
                contour.append(t_value)
            contour_values.append(contour)

        t_bounds = self.benchmarking_results["bias_adjusted_t"]
        return contour_values, critical_t, t_bounds

    def plot(
        self,
        plot_type="estimate",
        critical_value=None,
        x_limit=0.8,
        y_limit=0.8,
        num_points_per_contour=200,
        plot_size=(7, 7),
        contours_color="blue",
        critical_contour_color="red",
        label_fontsize=9,
        contour_linewidths=0.75,
        contour_linestyles="solid",
        contours_label_color="black",
        critical_label_color="red",
        unadjusted_estimate_marker="D",
        unadjusted_estimate_color="black",
        adjusted_estimate_marker="^",
        adjusted_estimate_color="red",
        legend_position=(1.6, 0.6),
    ):
        """
        Plots and summarizes the sensitivity bounds as a contour plot, as they vary with the partial R^2 of the unobserved confounder(s) with the treatment and the outcome
        Two types of plots can be generated, based on adjusted estimates or adjusted t-values
        X-axis: Partial R^2 of treatment and unobserved confounder(s)
        Y-axis: Partial R^2 of outcome and unobserved confounder(s)
        We also plot bounds on the partial R^2 of the unobserved confounders obtained from observed covariates

        :param plot_type: "estimate" or "t-value"
        :param critical_value: special reference value of the estimate or t-value that will be highlighted in the plot
        :param x_limit: plot's maximum x_axis value (default = 0.8)
        :param y_limit: plot's minimum y_axis value (default = 0.8)
        :param num_points_per_contour: number of points to calculate and plot each contour line (default = 200)
        :param plot_size: tuple denoting the size of the plot (default = (7,7))
        :param contours_color: color of contour line (default = blue)
                               String or array. If array, lines will be plotted with the specific color in ascending order.
        :param critical_contour_color: color of threshold line (default = red)
        :param label_fontsize: fontsize for labelling contours (default = 9)
        :param contour_linewidths: linewidths for contours (default = 0.75)
        :param contour_linestyles: linestyles for contours (default = "solid")
                                   See : https://matplotlib.org/3.5.0/gallery/lines_bars_and_markers/linestyles.html for more examples
        :param contours_label_color: color of contour line label (default = black)
        :param critical_label_color: color of threshold line label (default = red)
        :param unadjusted_estimate_marker: marker type for unadjusted estimate in the plot (default = 'D')
                                           See: https://matplotlib.org/stable/api/markers_api.html
        :parm unadjusted_estimate_color: marker color for unadjusted estimate in the plot (default = "black")
        :param adjusted_estimate_marker: marker type for bias adjusted estimates in the plot (default = '^')
        :parm adjusted_estimate_color: marker color for bias adjusted estimates in the plot (default = "red")
        :param legend_position:tuple denoting the position of the legend (default = (1.6, 0.6))
        """

        # Plotting the contour plot
        if plot_type == "estimate":
            critical_value = 0  # default value of estimate
        else:
            critical_value = 2  # default t-value (usual approx for 95% CI)

        fig, ax = plt.subplots(1, 1, figsize=plot_size)
        ax.set_title("Sensitivity contour plot of %s" % plot_type)
        ax.set_xlabel("Partial R^2 of confounder with treatment")
        ax.set_ylabel("Partial R^2 of confounder with outcome")

        for i in range(len(self.r2tu_w)):
            x = self.r2tu_w[i]
            y = self.r2yu_tw[i]
            if x > 0.8 or y > 0.8:
                x_limit = 0.99
                y_limit = 0.99
                break

        r2tu_w = np.arange(0.0, x_limit, x_limit / num_points_per_contour)
        r2yu_tw = np.arange(0.0, y_limit, y_limit / num_points_per_contour)

        unadjusted_point_estimate = None

        if plot_type == "estimate":
            contour_values, critical_value, bound_values = self.plot_estimate(r2tu_w, r2yu_tw)
            unadjusted_estimate = self.estimate
            unadjusted_point_estimate = unadjusted_estimate
        elif plot_type == "t-value":
            contour_values, critical_value, bound_values = self.plot_t(r2tu_w, r2yu_tw)
            unadjusted_t = self.t_stats
            unadjusted_point_estimate = unadjusted_t
        else:
            raise ValueError("Current plotting method only supports 'estimate' and 't-value' ")

        # Adding contours
        contour_plot = ax.contour(
            r2tu_w,
            r2yu_tw,
            contour_values,
            colors=contours_color,
            linewidths=contour_linewidths,
            linestyles=contour_linestyles,
        )
        ax.clabel(contour_plot, inline=1, fontsize=label_fontsize, colors=contours_label_color)

        # Adding threshold contour line
        contour_plot = ax.contour(
            r2tu_w,
            r2yu_tw,
            contour_values,
            colors=critical_contour_color,
            linewidths=contour_linewidths,
            levels=[critical_value],
        )
        ax.clabel(contour_plot, [critical_value], inline=1, fontsize=label_fontsize, colors=critical_label_color)

        # Adding unadjusted point estimate
        ax.scatter(
            [0],
            [0],
            marker=unadjusted_estimate_marker,
            color=unadjusted_estimate_color,
            label="Unadjusted({:1.2f})".format(unadjusted_point_estimate),
        )

        # Adding bounds to partial R^2 values for given strength of confounders
        for i in range(len(self.frac_strength_treatment)):
            frac_strength_treatment = self.frac_strength_treatment[i]
            frac_strength_outcome = self.frac_strength_outcome[i]
            if frac_strength_treatment == frac_strength_outcome:
                signs = str(round(frac_strength_treatment, 2))
            else:
                signs = str(round(frac_strength_treatment, 2)) + "/" + str(round(frac_strength_outcome, 2))
            label = (
                str(i + 1)
                + "  "
                + signs
                + " X "
                + str(self.original_benchmark_covariates)
                + " ({:1.2f}) ".format(bound_values[i])
            )
            ax.scatter(
                self.r2tu_w[i],
                self.r2yu_tw[i],
                color=adjusted_estimate_color,
                marker=adjusted_estimate_marker,
                label=label,
            )
            ax.annotate(str(i + 1), (self.r2tu_w[i] + 0.005, self.r2yu_tw[i] + 0.005))

        ax.legend(bbox_to_anchor=legend_position)
        plt.show()

    def __str__(self):
        s = "Sensitivity Analysis to Unobserved Confounding using R^2 paramterization\n\n"
        s += "Unadjusted Estimates of Treatment {0} :\n".format(self.original_treatment_name)
        s += "Coefficient Estimate : {0}\n".format(self.estimate)
        s += "Degree of Freedom : {0}\n".format(self.degree_of_freedom)
        s += "Standard Error : {0}\n".format(self.standard_error)
        s += "t-value : {0}\n".format(self.t_stats)
        s += "F^2 value : {0}\n\n".format(self.partial_f2)
        s += "Sensitivity Statistics : \n"
        s += "Partial R2 of treatment with outcome : {0}\n".format(self.stats["r2yt_w"])
        s += "Robustness Value : {0}\n\n".format(self.stats["robustness_value"])
        s += "Interpretation of results :\n"
        s += "Any confounder explaining less than {0}% percent of the residual variance of both the treatment and the outcome would not be strong enough to explain away the observed effect i.e bring down the estimate to 0 \n\n".format(
            round(self.stats["robustness_value"] * 100, 2)
        )
        s += "For a significance level of {0}%, any confounder explaining more than {1}% percent of the residual variance of both the treatment and the outcome would be strong enough to make the estimated effect not 'statistically significant'\n\n".format(
            self.significance_level * 100, round(self.stats["robustness_value_alpha"] * 100, 2)
        )
        s += "If confounders explained 100% of the residual variance of the outcome, they would need to explain at least {0}% of the residual variance of the treatment to bring down the estimated effect to 0\n".format(
            round(self.stats["r2yt_w"] * 100, 2)
        )
        return s



import copy

import pytest

from dowhy.causal_graph import CausalGraph
from dowhy.causal_identifier import AutoIdentifier, BackdoorAdjustment, EstimandType
from dowhy.causal_identifier.auto_identifier import EFFICIENT_METHODS
from tests.causal_identifiers.example_graphs_efficient import TEST_EFFICIENT_BD_SOLUTIONS


def test_identify_efficient_backdoor_algorithms():
    for example in TEST_EFFICIENT_BD_SOLUTIONS.values():
        G = CausalGraph(
            graph=example["graph_str"],
            treatment_name="X",
            outcome_name="Y",
            observed_node_names=example["observed_node_names"],
        )
        for method_name in EFFICIENT_METHODS:
            ident_eff = AutoIdentifier(
                estimand_type=EstimandType.NONPARAMETRIC_ATE,
                backdoor_adjustment=method_name,
                costs=example["costs"],
            )
            method_name_results = method_name.value.replace("-", "_")
            if example[method_name_results] is None:
                with pytest.raises(ValueError):
                    ident_eff.identify_effect(
                        G,
                        G.treatment_name,
                        G.outcome_name,
                        conditional_node_names=example["conditional_node_names"],
                    )
            else:
                results_eff = ident_eff.identify_effect(
                    G,
                    G.treatment_name,
                    G.outcome_name,
                    conditional_node_names=example["conditional_node_names"],
                )
                assert set(results_eff.get_backdoor_variables()) == example[method_name_results]


def test_fail_negative_costs_efficient_backdoor_algorithms():
    example = TEST_EFFICIENT_BD_SOLUTIONS["sr22_fig2_example_graph"]
    G = CausalGraph(
        graph=example["graph_str"],
        treatment_name="X",
        outcome_name="Y",
        observed_node_names=example["observed_node_names"],
    )
    mod_costs = copy.deepcopy(example["costs"])
    mod_costs[0][1]["cost"] = 0
    ident_eff = AutoIdentifier(
        estimand_type=EstimandType.NONPARAMETRIC_ATE,
        backdoor_adjustment=BackdoorAdjustment.BACKDOOR_MINCOST_EFFICIENT,
        costs=mod_costs,
    )

    with pytest.raises(Exception):
        ident_eff.identify_effect(
            G,
            G.treatment_name,
            G.outcome_name,
            conditional_node_names=example["conditional_node_names"],
        )


def test_fail_unobserved_cond_vars_efficient_backdoor_algorithms():
    example = TEST_EFFICIENT_BD_SOLUTIONS["sr22_fig2_example_graph"]
    G = CausalGraph(
        graph=example["graph_str"],
        treatment_name="X",
        outcome_name="Y",
        observed_node_names=example["observed_node_names"],
    )
    ident_eff = AutoIdentifier(
        estimand_type=EstimandType.NONPARAMETRIC_ATE,
        backdoor_adjustment=BackdoorAdjustment.BACKDOOR_MINCOST_EFFICIENT,
        costs=example["costs"],
    )
    mod_cond_names = copy.deepcopy(example["conditional_node_names"])
    mod_cond_names.append("U")
    with pytest.raises(Exception):
        ident_eff.identify_effect(
            G,
            G.treatment_name,
            G.outcome_name,
            conditional_node_names=mod_cond_names,
        )


def test_fail_multivar_treat_efficient_backdoor_algorithms():
    example = TEST_EFFICIENT_BD_SOLUTIONS["sr22_fig2_example_graph"]
    G = CausalGraph(
        graph=example["graph_str"],
        treatment_name=["X", "K"],
        outcome_name="Y",
        observed_node_names=example["observed_node_names"],
    )
    ident_eff = AutoIdentifier(
        estimand_type=EstimandType.NONPARAMETRIC_ATE,
        backdoor_adjustment=BackdoorAdjustment.BACKDOOR_MINCOST_EFFICIENT,
        costs=example["costs"],
    )
    with pytest.raises(Exception):
        ident_eff.identify_effect(
            G,
            G.treatment_name,
            G.outcome_name,
            conditional_node_names=example["conditional_node_names"],
        )


def test_fail_multivar_outcome_efficient_backdoor_algorithms():
    example = TEST_EFFICIENT_BD_SOLUTIONS["sr22_fig2_example_graph"]
    G = CausalGraph(
        graph=example["graph_str"],
        treatment_name="X",
        outcome_name=["Y", "R"],
        observed_node_names=example["observed_node_names"],
    )
    ident_eff = AutoIdentifier(
        estimand_type=EstimandType.NONPARAMETRIC_ATE,
        backdoor_adjustment=BackdoorAdjustment.BACKDOOR_MINCOST_EFFICIENT,
        costs=example["costs"],
    )
    with pytest.raises(Exception):
        ident_eff.identify_effect(
            G,
            G.treatment_name,
            G.outcome_name,
            conditional_node_names=example["conditional_node_names"],
        )
